#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <gtk/gtk.h>

#include "../include/string.h"
#include "../include/disk.h"
#include "../include/prochandle.h"

#include "guiutils.h"
#include "cdialog.h"

#include "edvtypes.h"
#include "edvcfg.h"
#include "edvobj.h"
#include "browser.h"
#include "browserdirtree.h"
#include "imbr.h"
#include "archiver.h"
#include "propwin.h"
#include "recbin.h"
#include "recbincontents.h"
#include "findwin.h"
#include "deviceswin.h"
#include "mimetypeswin.h"
#include "endeavour.h"
#include "edvop.h"
#include "edvcfglist.h"
#include "edvutils.h"
#include "config.h"


gint EDVDoNewBrowser(edv_core_struct *core_ptr);
gint EDVDoNewImbr(edv_core_struct *core_ptr);
gint EDVDoNewArchiverOpen(edv_core_struct *core_ptr, const gchar *arch_obj);
gint EDVDoNewArchiver(edv_core_struct *core_ptr);
gint EDVDoPropWinAndSwitchPage(
	edv_core_struct *core_ptr, edv_object_struct *obj,
	const gchar *tab_name
);
gint EDVDoPropWin(edv_core_struct *core_ptr, edv_object_struct *obj);

void EDVDoMapRecBinDeskIcon(edv_core_struct *core_ptr);
void EDVDoMapRecBin(edv_core_struct *core_ptr);
void EDVDoMapDevicesListWin(edv_core_struct *core_ptr);
void EDVDoMapMIMETypesListWin(edv_core_struct *core_ptr);

void EDVDoMapBrowserFindWin(
        edv_core_struct *core_ptr, edv_browser_struct *browser
);
void EDVDoMapImbrFindWin(
        edv_core_struct *core_ptr, edv_imbr_struct *imbr
);
void EDVDoMapArchiverFindWin(
        edv_core_struct *core_ptr, edv_archiver_struct *archiver
);
void EDVDoMapRecBinFindWin(
        edv_core_struct *core_ptr, edv_recbin_struct *recbin
);

void EDVDoUpdateDevicesPUList(edv_core_struct *core_ptr);
void EDVDoUpdateMIMETypeHintIndexes(edv_core_struct *core_ptr);
void EDVDoUpdateOpenWithPUList(edv_core_struct *core_ptr);

void EDVDoSyncDisks(edv_core_struct *core_ptr);

void EDVDoHelp(
	edv_core_struct *core_ptr, const gchar *topic, GtkWidget *toplevel
);
void EDVDoHelpAbout(edv_core_struct *core_ptr, GtkWidget *toplevel);



/* Help program argument types, specifies the type of argument it takes
 * (usually a file path or url).
 */
#define HELP_PROG_INPUT_TYPE_PATH       0
#define HELP_PROG_INPUT_TYPE_URL        1

/* Location of help programs, their respective arguments, and argument
 * type in sets of three pointers. The last set of three pointers in the
 * list should be all NULL's.
 */
#define HELP_PROG_LOCATIONS	{\
{ "/usr/bin/gnome-help-browser", "", HELP_PROG_INPUT_TYPE_URL }, \
{ "/usr/local/bin/gnome-help-browser", "", HELP_PROG_INPUT_TYPE_URL }, \
{ "/bin/gnome-help-browser", "", HELP_PROG_INPUT_TYPE_URL }, \
\
{ "/usr/bin/netscape", "", HELP_PROG_INPUT_TYPE_URL }, \
{ "/usr/local/bin/netscape", "", HELP_PROG_INPUT_TYPE_URL }, \
{ "/bin/netscape", "", HELP_PROG_INPUT_TYPE_URL }, \
\
{ "/usr/X11R6/bin/nxterm", "-e lynx", HELP_PROG_INPUT_TYPE_URL }, \
{ "/usr/X11R6/bin/xterm", "-e lynx", HELP_PROG_INPUT_TYPE_URL }, \
\
{ NULL, NULL, 0 } \
}

/* Structure for help program entry in the HELP_PROG_LOCATIONS list. */
typedef struct {
	gchar *prog;
	gchar *arg;
	gint arg_type;
} help_prog_struct;



/*
 *	Procedure to create a new browser and map it. The new browser
 *	will be appended to the core_ptr or an existing (unmapped)
 *	browser will be used.
 *
 *	Returns the index of the browser on the core structure or -1
 *	on error.
 */
gint EDVDoNewBrowser(edv_core_struct *core_ptr)
{
	gint i;
	edv_browser_struct *browser;


	if(core_ptr == NULL)
	    return(-1);

	/* Sanitize total. */
	if(core_ptr->total_browsers < 0)
	    core_ptr->total_browsers = 0;

	/* Get highest index and increase total. */
	i = core_ptr->total_browsers;
	core_ptr->total_browsers = i + 1;

	/* Allocate more pointers. */
	core_ptr->browser = (edv_browser_struct **)g_realloc(
	    core_ptr->browser,
	    core_ptr->total_browsers * sizeof(edv_browser_struct *)
	);
	if(core_ptr->browser == NULL)
	{
	    core_ptr->total_browsers = 0;
	    return(-1);
	}
	else
	{
	    core_ptr->browser[i] = NULL;
	}

	/* Create new browser. */
	core_ptr->browser[i] = browser = EDVBrowserNew(core_ptr);
	if(browser != NULL)
	{
	    gchar *startup_dir;
	    GtkCList *clist;


            EDVBrowserSetBusy(browser, TRUE);
	    GUIBlockInput(browser->toplevel, TRUE);

	    /* Need to map it immediatly after creation, before any
	     * timeouts check on it.
	     */
	    EDVBrowserMap(browser);

	    /* Get copy of start up directory. */
	    startup_dir = EDVCopyEvaluateInputPath(
		NULL,
		EDVCFGItemListGetValueS(
		    core_ptr->cfg_list, EDV_CFG_PARM_DIR_START_UP
		)
	    );

	    /* Get directory ctree and create initial toplevel nodes. */
	    clist = (GtkCList *)browser->directory_ctree;
	    if(clist != NULL)
	    {
		gtk_clist_freeze(clist);

		/* Create toplevel directories. */
		EDVBrowserDirTreeDoCreateToplevels(browser);

		gtk_clist_thaw(clist);


		/* Select startup directory, do not freeze. */
		EDVBrowserDirTreeDoSelectPath(browser, startup_dir);
	    }

            GUIBlockInput(browser->toplevel, FALSE);
	    EDVBrowserSetBusy(browser, FALSE);

	    /* Deallocate copy of start up directory. */
	    g_free(startup_dir);
	    startup_dir = NULL;

	    return(i);
	}
	else
	{
	    return(-1);
	}
}

/*
 *      Procedure to create a new image browser and map it. The new
 *	image browser will be appended to the core_ptr or an existing
 *	(unmapped)image browser will be used.
 *
 *      Returns the index of the browser on the core structure or -1
 *      on error.
 */
gint EDVDoNewImbr(edv_core_struct *core_ptr)
{
        gint i;
        edv_imbr_struct *imbr;


        if(core_ptr == NULL)
            return(-1);

        /* Sanitize total. */
        if(core_ptr->total_imbrs < 0)
            core_ptr->total_imbrs = 0;

        /* Get highest index and increase total. */
        i = core_ptr->total_imbrs;
        core_ptr->total_imbrs = i + 1;

        /* Allocate more pointers. */
        core_ptr->imbr = (edv_imbr_struct **)g_realloc(
            core_ptr->imbr,
            core_ptr->total_imbrs * sizeof(edv_imbr_struct *)
        );
        if(core_ptr->imbr == NULL)
        {
            core_ptr->total_imbrs = 0;
            return(-1);
        }
        else
        {
            core_ptr->imbr[i] = NULL;
        }

        /* Create new image browser. */
        core_ptr->imbr[i] = imbr = EDVImbrNew(core_ptr);
        if(imbr != NULL)
        {
            gchar *startup_dir;


	    EDVImbrSetBusy(imbr, TRUE);
            GUIBlockInput(imbr->toplevel, TRUE);

            /* Need to map it immediatly after creation, before any
             * timeouts check on it.
             */
            EDVImbrMap(imbr);

            /* Get copy of start up directory. */
            startup_dir = EDVCopyEvaluateInputPath(
                NULL,
                EDVCFGItemListGetValueS(
                    core_ptr->cfg_list, EDV_CFG_PARM_DIR_START_UP
                )
            );

            /* Switch to startup directory. */
	    EDVImbrSelectPath(imbr, startup_dir);

	    EDVImbrUpdateMenus(imbr);

            GUIBlockInput(imbr->toplevel, FALSE);
            EDVImbrSetBusy(imbr, FALSE);

            /* Deallocate copy of start up directory. */
            g_free(startup_dir);
            startup_dir = NULL;

            return(i);
        }
        else
        {
            return(-1);
        }
}

/*
 *      Procedure to create a new archiver and map it. The new
 *	archiver will be appended to the core_ptr or an existing
 *      (unmapped) archiver will be used.
 *
 *      Returns the index of the archiver on the core structure or -1
 *      on error.
 */
gint EDVDoNewArchiverOpen(edv_core_struct *core_ptr, const gchar *arch_obj)
{
        gint i;
        edv_archiver_struct *archiver;


        if(core_ptr == NULL)
            return(-1);

        /* Sanitize total. */
        if(core_ptr->total_archivers < 0)
            core_ptr->total_archivers = 0;

        /* Get highest index and increase total. */
        i = core_ptr->total_archivers;
        core_ptr->total_archivers = i + 1;

        /* Allocate more pointers. */
        core_ptr->archiver = (edv_archiver_struct **)g_realloc(
            core_ptr->archiver,
            core_ptr->total_archivers * sizeof(edv_archiver_struct *)
        );
        if(core_ptr->archiver == NULL)
        {
            core_ptr->total_archivers = 0;
            return(-1);
        }
        else
        {
            core_ptr->archiver[i] = NULL;
        }

        /* Create new archiver. */
        core_ptr->archiver[i] = archiver = EDVArchiverNew(core_ptr);
        if(archiver != NULL)
        {
            EDVArchiverSetBusy(archiver, TRUE);
            GUIBlockInput(archiver->toplevel, TRUE);

            /* Need to map it immediatly after creation, before any
             * timeouts check on it.
             */
            EDVArchiverMap(archiver);


	    if(arch_obj != NULL)
	    {
		/* Clear the contents clist and load the listing of the
		 * new archive specified by arch_obj.
                 */
                EDVArchiverSelectArchive(archiver, arch_obj);
	    }


            EDVArchiverUpdateMenus(archiver);

            GUIBlockInput(archiver->toplevel, FALSE);
            EDVArchiverSetBusy(archiver, FALSE);

            return(i);
        }
        else
        {
            return(-1);
        }
}

gint EDVDoNewArchiver(edv_core_struct *core_ptr)
{
	return(EDVDoNewArchiverOpen(core_ptr, NULL));
}


/*
 *	Procedure to create a property window and map it. The new property
 *	window will be appended to the core_ptr or an existing (unmapped)
 *	browser will be used.
 *
 *	Returns the index of the property window on the core structure or
 *	-1 on error.
 */
gint EDVDoPropWinAndSwitchPage(
        edv_core_struct *core_ptr, edv_object_struct *obj,
        const gchar *tab_name
)
{
        gint i;
        edv_propwin_struct *propwin;


        if(core_ptr == NULL)
            return(-1);

        /* Sanitize total. */
        if(core_ptr->total_propwins < 0)
            core_ptr->total_propwins = 0;

        /* Get highest index and increase total. */
        i = core_ptr->total_propwins;
        core_ptr->total_propwins = i + 1;

        /* Allocate more pointers. */
        core_ptr->propwin = (edv_propwin_struct **)g_realloc(
            core_ptr->propwin,
            core_ptr->total_propwins * sizeof(edv_propwin_struct *)
        );
        if(core_ptr->propwin == NULL)
        {
            core_ptr->total_propwins = 0;
            return(-1);
        }
        else
        {
            core_ptr->propwin[i] = NULL;
        }

        /* Create new property window. */
        core_ptr->propwin[i] = propwin = EDVPropWinNew(core_ptr, obj);
        if(propwin != NULL)
        {
	    GtkNotebook *notebook = GTK_NOTEBOOK(propwin->notebook);

	    /* Select tab? */
	    if((tab_name != NULL) && (notebook != NULL))
	    {
		gint page_num = -1, total_pages = 0;
		GList *glist = notebook->first_tab;


		/* Count total pages. */
		while(glist != NULL)
		{
		    total_pages++;
		    glist = glist->next;
		}

		/* Check which tab is to be switched to by matching its
		 * name.
		 */
		if(!strcasecmp(tab_name, "General"))
		    page_num = 0;
		else if(!strcasecmp(tab_name, "Link"))
                    page_num = 1;
                else if(!strcasecmp(tab_name, "Device"))
                    page_num = 1;
                else if(!strcasecmp(tab_name, "Device Node"))
                    page_num = 1;

		/* Matched page to be switched to? */
		if(page_num > -1)
		    gtk_notebook_set_page(notebook, page_num);
	    }

            /* Need to map it immediatly after creation, before any
             * timeouts check on it.
             */
            EDVPropWinMap(propwin);

            return(i);
        }
        else
        {
            return(-1);
        }
}

gint EDVDoPropWin(edv_core_struct *core_ptr, edv_object_struct *obj)
{
	return(EDVDoPropWinAndSwitchPage(core_ptr, obj, NULL));
}


/*
 *      Creates the recycle bin desktop icon as needed and maps it.
 */
void EDVDoMapRecBinDeskIcon(edv_core_struct *core_ptr)
{
	if(core_ptr == NULL)
	    return;

	if(core_ptr->recbin_deskicon == NULL)
	    core_ptr->recbin_deskicon = EDVRecBinDeskIconNew(core_ptr);
	EDVRecBinDeskIconMap(core_ptr->recbin_deskicon);
	EDVRecBinDeskIconUpdate(core_ptr->recbin_deskicon);
}

/*
 *	Creates the recycle bin window as needed and maps it.
 */
void EDVDoMapRecBin(edv_core_struct *core_ptr)
{
	GtkCList *clist;
        edv_recbin_struct *recbin;


        if(core_ptr == NULL)
            return;

        /* Get pointer to recycle bin window, create it as needed. */
	recbin = core_ptr->recbin;
        if(recbin == NULL)
            core_ptr->recbin = recbin = EDVRecBinNew(
                core_ptr
            );
        if(recbin == NULL)
            return;

	/* Need to map it immediatly after it is created so that it does
	 * not get deleted in the timeout callback when the recycle bin is
	 * updated further below.
	 */
        EDVRecBinMap(recbin);

	EDVRecBinSetBusy(recbin, TRUE);
	GUIBlockInput(recbin->toplevel, TRUE);

	clist = (GtkCList *)recbin->contents_clist;
	if(clist != NULL)
	{
	    gtk_clist_freeze(clist);
	    EDVRecBinContentsDoUpdate(recbin, TRUE);
	    gtk_clist_thaw(clist);
	}

	GUIBlockInput(recbin->toplevel, FALSE);
	EDVRecBinSetBusy(recbin, FALSE);

	EDVRecBinUpdateMenus(recbin);
}


/*
 *      Creates the devices list window as needed and maps it.
 */
void EDVDoMapDevicesListWin(edv_core_struct *core_ptr)
{
        edv_device_listwin_struct *lw;


        if(core_ptr == NULL)
            return;

        /* Get pointer to devices list window, create it as needed. */
        lw = core_ptr->device_listwin;
        if(lw == NULL)
            core_ptr->device_listwin = lw = EDVDevicesListWinNew(
                core_ptr
            );
        if(lw == NULL)
            return;

        EDVDevicesListWinMap(lw);

        EDVDevicesListWinFetchValues(lw);
}

/*
 *	Creates the MIME Types list window as needed and maps it.
 */
void EDVDoMapMIMETypesListWin(edv_core_struct *core_ptr)
{
	edv_mimetype_listwin_struct *lw;


	if(core_ptr == NULL)
	    return;

	/* Get pointer to MIME Types list window, create it as needed. */
	lw = core_ptr->mimetype_listwin;
	if(lw == NULL)
	    core_ptr->mimetype_listwin = lw = EDVMimeTypesListWinNew(
		core_ptr
	    );
	if(lw == NULL)
	    return;

	EDVMimeTypesListWinMap(lw);

	EDVMimeTypesListWinFetchValues(lw);
}


/*
 *	Maps the find window relative to the given browser.
 */
void EDVDoMapBrowserFindWin(
        edv_core_struct *core_ptr, edv_browser_struct *browser
)
{
	gint i, browser_num = -1;
	edv_findwin_struct *fw;


	if((core_ptr == NULL) || (browser == NULL))
	    return;

	/* Get pointer to find window and create the find window if it has
	 * not been created yet.
	 */
	fw = core_ptr->findwin;
	if(fw == NULL)
	    core_ptr->findwin = fw = EDVFindWinNew(core_ptr);
	if(fw == NULL)
	    return;

	/* Find browser index on core structure that matches the pointer
	 * of the given browser and record as browser_num.
	 */
	for(i = 0; i < core_ptr->total_browsers; i++)
	{
	    if(core_ptr->browser[i] == browser)
	    {
		browser_num = i;
		break;
	    }
	}

	/* Set reference window for find window, only one input may be
	 * non-negative.
	 */
	EDVFindWinSetReferenceWindow(
	    fw,
	    browser_num,
	    -1,
	    -1,
	    -1
	);

	/* Set initial location of find window to match the current
	 * location of the given browser.
	 */
	EDVFindWinSetLocation(
	    fw,
	    EDVBrowserCurrentLocation(browser),
	    FALSE
	);

	EDVFindWinUpdateMenus(fw);
	EDVFindWinMap(fw);
}

/*
 *      Maps the find window relative to the given image browser.
 */
void EDVDoMapImbrFindWin(
        edv_core_struct *core_ptr, edv_imbr_struct *imbr
)
{
        gint i, imbr_num = -1;
        edv_findwin_struct *fw;


        if((core_ptr == NULL) || (imbr == NULL))
            return;

        /* Get pointer to find window and create the find window if it has
         * not been created yet.
         */
        fw = core_ptr->findwin;
        if(fw == NULL)
            core_ptr->findwin = fw = EDVFindWinNew(core_ptr);
        if(fw == NULL)
            return;

        /* Find image browser index on core structure that matches the
	 * pointer of the given image browser and record as imbr_num.
         */
        for(i = 0; i < core_ptr->total_imbrs; i++)
        {
            if(core_ptr->imbr[i] == imbr)
            {
                imbr_num = i;
                break;
            }
        }

        /* Set reference window for find window, only one input may be
         * non-negative.
         */
        EDVFindWinSetReferenceWindow(
            fw,
            -1,
	    imbr_num,
            -1,
	    -1
        );

        /* Set initial location of find window to match the current
         * location of the given image browser.
         */
        EDVFindWinSetLocation(
            fw,
            EDVImbrCurrentLocation(imbr),
            FALSE
        );

        EDVFindWinUpdateMenus(fw);
        EDVFindWinMap(fw);
}

/*
 *	Maps the find window relative to the given archiver.
 */
void EDVDoMapArchiverFindWin(
        edv_core_struct *core_ptr, edv_archiver_struct *archiver
)
{
	gint i, archiver_num = -1;
        edv_findwin_struct *fw;


        if((core_ptr == NULL) || (archiver == NULL))
            return;

        /* Get pointer to find window and create the find window if it has
         * not been created yet.
         */
        fw = core_ptr->findwin;
        if(fw == NULL)
            core_ptr->findwin = fw = EDVFindWinNew(core_ptr);
        if(fw == NULL)
            return;

        /* Find archiver index on core structure that matches the
         * pointer of the given archiver and record as archiver_num.
         */
        for(i = 0; i < core_ptr->total_archivers; i++)
        {
            if(core_ptr->archiver[i] == archiver)
            {
                archiver_num = i;
                break;
            }
        }

        /* Set reference window for find window, only one input may be
         * non-negative.
         */
        EDVFindWinSetReferenceWindow(
            fw,
            -1,
            -1,
            -1,
            archiver_num
        );

        /* Set initial location of find window to match the current
         * location of the given archiver.
         */
        EDVFindWinSetLocation(
            fw,
            EDVArchiverCurrentLocation(archiver),
            FALSE
        );

        EDVFindWinUpdateMenus(fw);
        EDVFindWinMap(fw);
}

/*
 *      Maps the find window relative to the given recycle bin.
 */
void EDVDoMapRecBinFindWin(
        edv_core_struct *core_ptr, edv_recbin_struct *recbin
)
{
	const gchar *recycled_index_file;
        edv_findwin_struct *fw;


        if((core_ptr == NULL) || (recbin == NULL))
            return;

        /* Get pointer to find window and create the find window if it has
         * not been created yet.
         */
        fw = core_ptr->findwin;
        if(fw == NULL)
            core_ptr->findwin = fw = EDVFindWinNew(core_ptr);
        if(fw == NULL)
            return;

        /* Set reference window for find window, only one input may be
         * non-negative.
         */
        EDVFindWinSetReferenceWindow(
            fw,
            -1,
            -1,
	    0,		/* First and only recycle bin. */
	    -1
        );


        /* Get path to recycled objects index file. */
        recycled_index_file = EDVCFGItemListGetValueS(
            core_ptr->cfg_list, EDV_CFG_PARM_FILE_RECYCLED_INDEX
        );

        /* Set initial location of find window to be the recycled
	 * objects directory.
         */
        EDVFindWinSetLocation(
            fw,
            GetParentDir(recycled_index_file),
            FALSE
        );

        EDVFindWinUpdateMenus(fw);
        EDVFindWinMap(fw);
}

/*
 *	Deletes all items in the devices popup list widget (if any) and
 *	adds the current devices on the core structure to the devices
 *	popup list.
 *
 *	If the devices popup list does not exist on the core structure
 *	then it will be created. So this function can also be used to
 *	initialize it.
 */
void EDVDoUpdateDevicesPUList(edv_core_struct *core_ptr)
{
	gint i;
	const gchar *dev_name;
	edv_device_struct *dev_ptr;
	pulist_struct *pulist;


	if(core_ptr == NULL)
	    return;

	/* Get devices popup list from core structure, create it as 
	 * needed.
	 */
	pulist = core_ptr->devices_pulist;
	if(pulist == NULL)
	{
	    /* Popup list does not exist, create it. */
	    core_ptr->devices_pulist = pulist = PUListNew();
	}
	if(pulist == NULL)
	    return;

	/* Delete all existing items from the popup list. */
	PUListClear(pulist);

	/* Add current devices list to popup list. */
	for(i = 0; i < core_ptr->total_devices; i++)
	{
	    dev_ptr = core_ptr->device[i];
	    if(dev_ptr == NULL)
		continue;

	    /* Skip devices who are marked as unlisted, in which case 
	     * they should not be listed on the devices popup list.
	     */
	    if(dev_ptr->unlisted)
		continue;


	    /* Realize device, loading icons as needed. */
	    EDVDeviceRealize(dev_ptr, FALSE);

	    /* Get device name. */
	    dev_name = dev_ptr->name;
	    if((dev_name == NULL) && (dev_ptr->device_path != NULL))
	    {
		dev_name = strrchr(dev_ptr->device_path, DIR_DELIMINATOR);
		if(dev_name == NULL)
		    dev_name = dev_ptr->device_path;
		else
		    dev_name++;
	    }
	    if(dev_name == NULL)
		dev_name = "(null)";

	    /* Add new item to devices popup list. */
	    PUListAddItemPixText(
		pulist, dev_name,
		dev_ptr->small_pixmap[0], dev_ptr->small_mask[0],
                (gpointer)i, NULL
	    );
	}
}

/*
 *	Updates the MIME Type class hint index values on the given core
 *	structure to the MIME Types list.
 */
void EDVDoUpdateMIMETypeHintIndexes(edv_core_struct *core_ptr)
{
	gint i;
	edv_mimetype_struct *mt_ptr;


	if(core_ptr == NULL)
	    return;

	/* Reset all indexes for all MIME Type classes. */
	core_ptr->mimetype_system_index_first = -1;
	core_ptr->mimetype_system_index_last = -1;
	core_ptr->mimetype_format_index_first = -1;
	core_ptr->mimetype_format_index_last = -1;
	core_ptr->mimetype_program_index_first = -1;
	core_ptr->mimetype_program_index_last = -1;
	core_ptr->mimetype_unique_index_first = -1;
	core_ptr->mimetype_unique_index_last = -1;

	/* Iterate through all MIME Types and update each class hint
	 * index.
	 */
	for(i = 0; i < core_ptr->total_mimetypes; i++)
	{
	    mt_ptr = core_ptr->mimetype[i];
	    if(mt_ptr == NULL)
		continue;

	    switch(mt_ptr->mt_class)
	    {
	      case EDV_MIMETYPE_CLASS_SYSTEM:
		if(core_ptr->mimetype_system_index_first < 0)
		    core_ptr->mimetype_system_index_first = i;
		core_ptr->mimetype_system_index_last = i;
		break;

              case EDV_MIMETYPE_CLASS_FORMAT:
                if(core_ptr->mimetype_format_index_first < 0)
                    core_ptr->mimetype_format_index_first = i;
                core_ptr->mimetype_format_index_last = i;
                break;

              case EDV_MIMETYPE_CLASS_PROGRAM:
                if(core_ptr->mimetype_program_index_first < 0)
                    core_ptr->mimetype_program_index_first = i;
                core_ptr->mimetype_program_index_last = i;
                break;

              case EDV_MIMETYPE_CLASS_UNIQUE:
                if(core_ptr->mimetype_unique_index_first < 0)
                    core_ptr->mimetype_unique_index_first = i;
                core_ptr->mimetype_unique_index_last = i;
                break;
	    }
	}

}

/*
 *      Deletes all items in the `open with' popup list widget (if any) and
 *      adds the current MIME Types of class application on the core
 *	structure to the `open with' popup list.
 *
 *      If the `open with' popup list does not exist on the core structure
 *      then it will be created. So this function can also be used to
 *      initialize it.
 */
void EDVDoUpdateOpenWithPUList(edv_core_struct *core_ptr)
{
        gint i;
        edv_mimetype_struct *mt_ptr;
        pulist_struct *pulist;


        if(core_ptr == NULL)
            return;

        /* Get `open with' popup list from core structure, create it as
         * needed.
         */
        pulist = core_ptr->openwith_pulist;
        if(pulist == NULL)
        {
            /* Popup list does not exist, create it. */
            core_ptr->openwith_pulist = pulist = PUListNew();
        }
        if(pulist == NULL)
            return;

        /* Delete all existing items from the popup list. */
        PUListClear(pulist);

        /* Add current MIME Types of class application to the popup list. */
        for(i = 0; i < core_ptr->total_mimetypes; i++)
        {
            mt_ptr = core_ptr->mimetype[i];
            if(mt_ptr == NULL)
                continue;

	    if(mt_ptr->type == NULL)
		continue;

	    /* Skip MIME Type if it is not of class application. */
	    if(mt_ptr->mt_class != EDV_MIMETYPE_CLASS_PROGRAM)
		continue;

            /* Realize MIME Type, loading icons as needed. */
            EDVMimeTypeRealize(mt_ptr, FALSE);

            /* Add new item to `open with' popup list. */
            PUListAddItemPixText(
                pulist,
		(mt_ptr->description != NULL) ?
		    mt_ptr->description : mt_ptr->type,
                mt_ptr->small_pixmap[0], mt_ptr->small_mask[0],
		(gpointer)i, NULL
            );
	}
}


/*
 *	Procedure to sync all data to disk by calling sync().
 */
void EDVDoSyncDisks(edv_core_struct *core_ptr)
{
	static gbool reenterent = FALSE;

	if(core_ptr == NULL)
	    return;

	if(reenterent)
	    return;
	else
	    reenterent = TRUE;

	sync();

	reenterent = FALSE;
}


/*
 *	Help calling nexus.
 *
 *	If topic is NULL then it is assumed to display the help contents.
 */
void EDVDoHelp(
	edv_core_struct *core_ptr, const gchar *topic, GtkWidget *toplevel
)
{
	gint i;
	const gchar *data_dir, *help_file, *help_path;
	help_prog_struct *hp_ptr;
	help_prog_struct help_prog_list[] = HELP_PROG_LOCATIONS;


	if(core_ptr == NULL)
	    return;

	/* If no topic is given, then assume it means the help contents are
	 * to be displayed.
	 */
	if(topic == NULL)
	    topic = "Contents";

	/* Get global data directory in which the help files are assumed
	 * to be located in.
	 */
	data_dir = EDVCFGItemListGetValueS(
	    core_ptr->cfg_list, EDV_CFG_PARM_DIR_GLOBAL
	);
	if(data_dir == NULL)
	    return;

	/* See which help file to load based on the given topic name. */
	if(!strcasecmp(topic, "Contents"))
	    help_file = "help/index.html";
        else if(!strcasecmp(topic, "Keys List"))
            help_file = "help/keys_list.html";
        else if(!strcasecmp(topic, "MIME Types"))
            help_file = "help/mime_types.html";
        else if(!strcasecmp(topic, "Devices"))
            help_file = "help/devices.html";
        else if(!strcasecmp(topic, "Common Operations"))
            help_file = "help/common_operations.html";
        else
	    help_file = NULL;

	if(help_file == NULL)
	    help_path = NULL;
	else
	    help_path = PrefixPaths(data_dir, help_file);

	/* No help file matched for the given topic? */
	if(help_path == NULL)
	{
	    gchar *buf = g_strdup_printf(
		"Unable to find internal help topic `%s'.",
		topic
	    );

	    CDialogSetTransientFor(toplevel);
	    CDialogGetResponse(
		"No Help Available", buf, NULL,
		CDIALOG_ICON_ERROR,
		CDIALOG_BTNFLAG_OK,
		CDIALOG_BTNFLAG_OK
	    );
	    CDialogSetTransientFor(NULL);

	    g_free(buf);

	    return;
	}


	/* Search for a help program to use. */
	i = 0;
	hp_ptr = NULL;
	while(help_prog_list[i].prog != NULL)
	{
	    const gchar *help_prog = help_prog_list[i].prog;
	    if(!access(help_prog, X_OK))
	    {
		hp_ptr = &help_prog_list[i];
		break;
	    }
	    i++;
	}

	/* Found a help program? */
	if(hp_ptr != NULL)
	{
            pid_t p;
            gchar *cmd = NULL;

	    /* Format command string. */
	    cmd = strcatalloc(cmd, hp_ptr->prog);
	    if(hp_ptr->arg != NULL)
	    {
		cmd = strcatalloc(cmd, " ");
		cmd = strcatalloc(cmd, hp_ptr->arg);
	    }
	    switch(hp_ptr->arg_type)
	    {
	      case HELP_PROG_INPUT_TYPE_URL:
		cmd = strcatalloc(cmd, " file://");
		cmd = strcatalloc(cmd, help_path);
		break;

	      default:	/* HELP_PROG_INPUT_TYPE_PATH */
		cmd = strcatalloc(cmd, " ");
		cmd = strcatalloc(cmd, help_path);
		break;
	    }

	    /* Execute command. */
	    p = Exec(cmd);

	    /* Execution of help viewer failed? */
	    if(p == 0)
	    {
		gchar *buf = g_strdup_printf(
"Unable to execute help viewer command:\n\
\n\
    %s\n",
		    cmd
		);

		CDialogSetTransientFor(toplevel);
		CDialogGetResponse(
"Help Viewer Execution Failed",
		    buf,
"An error was encountered while attempting to run the help\n\
viewer to display the requested document. Please verify that\n\
the help viewer and the requested document are both\n\
installed properly.",
		    CDIALOG_ICON_ERROR,
		    CDIALOG_BTNFLAG_OK | CDIALOG_BTNFLAG_HELP,
		    CDIALOG_BTNFLAG_OK
		);

		g_free(buf);
	    }

	    /* Deallocate command string. */
	    g_free(cmd);
            cmd = NULL;
	}
	else
	{
            CDialogSetTransientFor(toplevel);
            CDialogGetResponse(
		"No Help Viewer",
"Unable to find a help viewer to display help documentation",
		NULL,
                CDIALOG_ICON_ERROR,
                CDIALOG_BTNFLAG_OK,
                CDIALOG_BTNFLAG_OK
            );
            CDialogSetTransientFor(NULL);
	}
}

/*
 *	Maps the about window.
 */
void EDVDoHelpAbout(edv_core_struct *core_ptr, GtkWidget *toplevel)
{
        if(core_ptr == NULL)
            return;

        if(core_ptr->about_dialog == NULL)
            core_ptr->about_dialog = AboutDialogNew(core_ptr);
        AboutDialogMap(core_ptr->about_dialog);
}
