#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <gtk/gtk.h>

#include "../include/string.h"
#include "../include/disk.h"
#include "../include/fio.h"

#include "guiutils.h"
#include "cdialog.h"

#include "edvtypes.h"
#include "edvcfg.h"
#include "edvhistory.h"
#include "edvmimetypes.h"
#include "browser.h"
#include "endeavour.h"
#include "edvutils.h"
#include "edvcb.h"
#include "edvcfglist.h"
#include "config.h"



void EDVSimplifyPath(gchar *path);
gbool EDVIsParentPath(const gchar *parent, const gchar *child);
gbool EDVIsExtension(const gchar *path, const gchar *ext);

gchar *EDVCopyShortenPath(const gchar *path, gint max);
gchar *EDVCopyEvaluateInputPath(
	const gchar *parent, const gchar *input_path
);
gchar *EDVTmpName(const gchar *dir);

void EDVGetDeviceNumbers(dev_t rdev, gint *major, gint *minor);
dev_t EDVFormatDeviceNumbers(gint major, gint minor);

GdkCursor *EDVGetCursor(
	edv_core_struct *core_ptr, gint cursor_code
);

GtkCTreeNode *EDVNodeGetParent(GtkCTreeNode *node);
GtkCTreeNode *EDVNodeGetChild(GtkCTreeNode *node);
GtkCTreeNode *EDVNodeGetSibling(GtkCTreeNode *node);
GtkCTreeNode *EDVNodeGetToplevel(GtkCTree *ctree);

gint EDVCListGetSelectedLast(
	GtkCList *clist, GtkCListRow **row_ptr_rtn
);
GtkCTreeNode *EDVCTreeGetSelectedLast(
	GtkCTree *ctree, GtkCTreeRow **row_ptr_rtn
);

gbool EDVNodeGetIndex(
        GtkCTree *ctree, GtkCTreeNode *node,
        gint *row_rtn, gint *column_rtn
);
GtkCTreeNode *EDVNodeGetByCoordinates(
        GtkCTree *ctree, gint x, gint y
);
void EDVScrollCListToPosition(
	GtkCList *clist, gfloat hpos, gfloat vpos
);

gbool EDVObjectNameNotationOK(const gchar *name);
gbool EDVCheckObjectHidden(
	edv_core_struct *core_ptr, edv_object_struct *object
);

gbool EDVCheckWriteProtect(edv_core_struct *core_ptr, gbool verbose);

gbool EDVCheckUIDIsOwner(gint effective_uid, gint owner_id);
gbool EDVCheckIsOwner(edv_core_struct *core_ptr, gint owner_id);

gbool EDVCheckObjectAccessable(
	edv_core_struct *core_ptr, edv_object_struct *object
);
gbool EDVCheckObjectReadable(
        edv_core_struct *core_ptr, edv_object_struct *object
);
gbool EDVCheckObjectWriteable(
        edv_core_struct *core_ptr, edv_object_struct *object
);
gbool EDVCheckObjectExecutable(
        edv_core_struct *core_ptr, edv_object_struct *object
);

gbool EDVCheckImlibImage(
	edv_core_struct *core_ptr, const gchar *path
);
gbool EDVCheckEDVArchiverArchive(
        edv_core_struct *core_ptr, const gchar *path
);

gint EDVMatchObjectIcon(
	edv_device_struct **device, gint total_devices,
	edv_mimetype_struct **mimetype, gint total_mimetypes,
	gint type,		/* One of EDV_OBJECT_TYPE_*. */
	const gchar *path,	/* Full path or just name. */
        gbool link_valid, guint permissions,
        gint icon_size,         /* 0 = small, 1 = medium, 2 = large. */
	GdkPixmap **pixmap_closed, GdkBitmap **mask_closed,
        GdkPixmap **pixmap_opened, GdkBitmap **mask_opened,
        GdkPixmap **pixmap_extended, GdkBitmap **mask_extended
);
gint EDVMatchObjectTypeString(
        edv_mimetype_struct **mimetype, gint total_mimetypes,
        gint type,              /* One of EDV_OBJECT_TYPE_*. */
        guint permissions,
        const gchar *path,      /* Full path or just name. */
	gchar **type_str
);

void EDVShowFileInCDialog(
        const gchar *path, const gchar *title, gint cdialog_icon,
	GtkWidget *toplevel
);

void EDVEntrySetDND(edv_core_struct *core_ptr, GtkWidget *w);

void EDVAppendHistory(
	edv_core_struct *core_ptr,
	gint type,		/* One of EDV_HISTORY_*. */
	gulong time_start,	/* Time the operation first started. */
	gulong time_end,	/* Time the operation ended. */
	gint status,
	const gchar *src_path, const gchar *tar_path,
	const gchar *comments
);


#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))



/*
 *	Simplifies the path by removing any tailing deliminator characters
 *	and reducing any occurances of "/.." in the given path.
 */
void EDVSimplifyPath(gchar *path)
{
	gbool is_absolute = FALSE;


	if(path == NULL)
	    return;

	if(*path == '\0')
	    return;

	if(*path == DIR_DELIMINATOR)
	    is_absolute = TRUE;

	/* Remove tailing deliminator characters if any. */
	StripPath(path);

	/* Reduce all occurances of "/.." in the path. */
	SimplifyPath(path);

	/* If the path was originally an absolute path and now its
	 * become an empty string then set path as toplevel.
	 */
	if(is_absolute && (*path == '\0'))
	    strcpy(path, "/");
}


/*
 *	Returns TRUE if the given parent path is a parent or grand
 *	parent of the given child path.
 *
 *	Will also return TRUE if parent path and child paths are
 *	identical.
 *
 *	Both paths must be absolute and simplified.
 */
gbool EDVIsParentPath(const gchar *parent, const gchar *child)
{
	gbool status;
	gint len;
	gchar *lparent, *lchild;
	gchar delim_str[2];

	
	if((parent == NULL) || (child == NULL))
	    return(FALSE);

	/* Both paths must be absolute. */
	if(!ISPATHABSOLUTE(parent) || !ISPATHABSOLUTE(child))
	    return(FALSE);

#define DO_FREE_LOCALS	\
{ \
 g_free(lparent); \
 lparent = NULL; \
\
 g_free(lchild); \
 lchild = NULL; \
}
	/* Make coppies of the parent and child paths. */
	lparent = g_strdup(parent);
	lchild = g_strdup(child);
	if((lparent == NULL) || (lchild == NULL))
	{
	    DO_FREE_LOCALS
	    return(FALSE);
	}


	/* Need to tack on a tailing DIR_DELIMINATOR character to the
	 * parent path for proper prefix matching below.
	 */
	delim_str[0] = DIR_DELIMINATOR;
	delim_str[1] = '\0';

	len = strlen(lparent);
	if(len > 0)
	{
	    if(lparent[len - 1] != DIR_DELIMINATOR)
		lparent = strcatalloc(lparent, delim_str);
	    if(lparent == NULL)
	    {
		DO_FREE_LOCALS
		return(FALSE);
	    }
        }

        len = strlen(lchild);
        if(len > 0)
        {
            if(lchild[len - 1] != DIR_DELIMINATOR)
                lchild = strcatalloc(lchild, delim_str);
            if(lchild == NULL)
            {
		DO_FREE_LOCALS
                return(FALSE);
            }
        }

	/* Check if the lparent is a prefix of lchild, if it is then
	 * it suggest that it is a parent or grand parent of the child
	 * path.
	 */
	if(strpfx(lchild, lparent))
	    status = TRUE;
	else
	    status = FALSE;

	/* Deallocate coppies of child and parent paths. */
	DO_FREE_LOCALS
#undef DO_FREE_LOCALS

	return(status);
}

/*
 *	Checks if atleast one of the given list of space separated extensions
 *	ext is the given path's extensions.
 */
gbool EDVIsExtension(const gchar *path, const gchar *ext)
{
	const gchar *ext_ptr, *ext_end_ptr, *path_end, *path_end_ptr;


	if((path == NULL) || (ext == NULL))
	    return(FALSE);

	/* Given path and extensions list cannot be empty strings. */
	if((*path == '\0') || (*ext == '\0'))
	    return(FALSE);

	/* Get pointer to end of the given path string (at the terminating
	 * null character).
	 */
	path_end = path;
	while(*path_end != '\0')
	    path_end++;

	/* Iterate through the given space separated list of extensions.
	 * Copying each individual extension string segment to cur_ext
	 * string buffer and then comparing to see if that is the
	 * postfix of the given path matches the extension.
	 */
	ext_ptr = ext;
	while(ISBLANK(*ext_ptr))
	    ext_ptr++;
	while(*ext_ptr != '\0')
	{
	    /* Seek ext_end_ptr to the end of the current extension or
	     * or extension string, then seek back one character.
	     */
	    ext_end_ptr = ext_ptr;
	    while(!ISBLANK(*ext_end_ptr) && (*ext_end_ptr != '\0'))
		ext_end_ptr++;
	    ext_end_ptr--;

	    /* Seek path_end_ptr to the end of the given path then seek
	     * back one character.
	     */
	    path_end_ptr = path_end - 1;

	    /* Compare the current extension with the postfix of the
	     * given path (case insensitive).
	     */
	    while((ext_end_ptr >= ext_ptr) &&
                  (path_end_ptr >= path)
	    )
	    {
		if(toupper(*ext_end_ptr) != toupper(*path_end_ptr))
		    break;

		ext_end_ptr--;
		path_end_ptr--;
	    }
	    /* If the ext_end_ptr iterated all the way back and under
	     * ext_ptr then that implies that no characters of the
	     * current extension are different with the postfix of the
	     * given path.
	     */
	    if(ext_end_ptr < ext_ptr)
		return(TRUE);

	    /* Seek to next extension (if any). */
	    while(!ISBLANK(*ext_ptr) && (*ext_ptr != '\0'))
		ext_ptr++;
	    while(ISBLANK(*ext_ptr))
		ext_ptr++;
	}

	return(FALSE);
}


/*
 *	Returns a dynamically allocated path that contains no more
 *	than max characters.
 *
 *	If the given path is longer than max then the shortened
 *	return stirng will contain "..." prefixed to it.
 */
gchar *EDVCopyShortenPath(const gchar *path, gint max)
{
	gint len;


	if(path == NULL)
	    return(NULL);

	len = strlen(path);
	if((len > max) && (max > 3))
	{
	    /* Need to shorten string. */
	    gint i = len - max + 3;

	    return(g_strdup_printf(
		"...%s", &path[i]
	    ));
	}
	else
	{
	    return(g_strdup(path));   
	}
}

/*
 *	Returns a dynamically allocated path evaluated (all variables
 *	and special notations substituted) based on the given input_path.
 *
 *	If input_path (after evaluation) is not an absolute path then the
 *	given parent path (if it is not NULL) will be prefixed to it.
 *
 *	The final path will be simplified.
 *
 *	The calling function must deallocate the returned pointer.
 */
gchar *EDVCopyEvaluateInputPath(
        const gchar *parent, const gchar *input_path
)
{
	gchar *new_path;


	if(input_path == NULL)
	    return(NULL);

	/* Check for special prefix substitutions for the input path. */
	if(strcasepfx(input_path, "~"))
	{
	    /* Prefix the value from the HOME environment variable to
	     * the input path and generate the new path.
	     */
	    const gchar *cstrptr = getenv("HOME");
	    const gchar *cstrptr2 = input_path + strlen("~");

	    /* If home directory environment variable value was not set
	     * then assume toplevel.
	     */
	    if(cstrptr == NULL)
		cstrptr = "/";

	    /* Seek input path after the special prefix and past any
	     * deliminator characters (deliminator character will be added
	     * when the new path is generated).
	     */
	    while(*cstrptr2 == DIR_DELIMINATOR)
		cstrptr2++;

	    /* Generate new path. */
	    new_path = g_strdup_printf(
		"%s%c%s",
		cstrptr, DIR_DELIMINATOR, cstrptr2
	    );
	}
	else
	{
	    /* If the given input pat his an absolute path then do not
	     * prefix the given parent to it.
	     */
	    if(ISPATHABSOLUTE(input_path))
	    {
		new_path = g_strdup(input_path);
	    }
	    else
	    {
		new_path = g_strdup_printf(
		    "%s%c%s",
		    (parent != NULL) ? parent : "/",
		    DIR_DELIMINATOR,
		    input_path
		);
	    }
	}

	/* Begin simplifying the generated path. */

        /* Simplify path, reducing occurances of "../". */
        SimplifyPath(new_path);

        /* Remove tailing deliminators. */
        StripPath(new_path);

	return(new_path);
}

/*
 *	Generates a unique file name in the directory specified by dir
 *	that is gauranteed to not exist.
 *
 *	The calling function must deallocate the returned pointer.
 */
gchar *EDVTmpName(const gchar *dir)
{
	guint i;
	gchar *path;
	struct stat lstat_buf;


	if(dir == NULL)
	    dir = "/tmp";

	/* Try and find a non-existant path by using a counter as the
	 * object name postfixed to the given directory.
	 */
	for(i = 0; i < (guint)-1; i++)
	{
	    path = g_strdup_printf("%s/%i", dir, i);

	    /* Does not exist? */
	    if(lstat(path, &lstat_buf))
		return(path);

	    /* Path exists, deallocate it and try again. */
	    g_free(path);
	}

	/* No possible path found, give up. */
	return(NULL);
}


/*
 *	Parses and returns the major and minor numbers for the device
 *	given by rdev which should be obtained from a prior call to
 *	*stat() in structure struct stat stat_buf.st_rdev.
 */
void EDVGetDeviceNumbers(dev_t rdev, gint *major, gint *minor)
{
	if(major != NULL)
	    *major = ((rdev >> 8) & 0xff);
	if(minor != NULL)
	    *minor = ((rdev >> 0) & 0xff);
}

/*
 *	Returns a formatted rdev value formatted from the given device
 *	major and minor numbers.
 */
dev_t EDVFormatDeviceNumbers(gint major, gint minor)
{
	return(((dev_t)major << 8) | ((dev_t)minor << 0));
}


/*
 *	Returns a GdkCursor specified by cursor_code on the given core
 *	structure.
 */
GdkCursor *EDVGetCursor(
	edv_core_struct *core_ptr, gint cursor_code
)
{
	gint i;
	edv_cursor_struct *cur;


	if(core_ptr == NULL)
	    return(NULL);

	for(i = 0; i < core_ptr->total_cursors; i++)
	{
	    cur = core_ptr->cursor[i];
	    if(cur == NULL)
		continue;

	    if(cur->code == cursor_code)
		return(cur->cursor);
	}

	return(NULL);
}


/*
 *	Returns the parent of the given node or NULL if there isn't
 *	one.
 */
GtkCTreeNode *EDVNodeGetParent(GtkCTreeNode *node)
{
        GtkCTreeRow *row_ptr;


        if(node == NULL)
            return(NULL);

        row_ptr = GTK_CTREE_ROW(node);
        if(row_ptr == NULL)
            return(NULL);

        return(row_ptr->parent);
}

/*
 *	Returns the first child of the given node or NULL if there
 *	isn't any.
 */
GtkCTreeNode *EDVNodeGetChild(GtkCTreeNode *node)
{
	GtkCTreeRow *row_ptr;


	if(node == NULL)
	    return(NULL);

	row_ptr = GTK_CTREE_ROW(node);
	if(row_ptr == NULL)
	    return(NULL);

	return(row_ptr->children);
}

/*
 *      Returns the next sibling of the given node or NULL if there
 *      isn't any.
 */
GtkCTreeNode *EDVNodeGetSibling(GtkCTreeNode *node)
{
        GtkCTreeRow *row_ptr;


        if(node == NULL)
            return(NULL);

        row_ptr = GTK_CTREE_ROW(node);
        if(row_ptr == NULL)
            return(NULL);

        return(row_ptr->sibling);
}

/*
 *	Returns the first toplevel node on the given GtkCTree widget
 *	or NULL if there isn't any.
 */
GtkCTreeNode *EDVNodeGetToplevel(GtkCTree *ctree)
{
	if(ctree == NULL)
	    return(NULL);

	return(gtk_ctree_node_nth(ctree, 0));
}


/*
 *	Returns the index of the last selected row on the given clist.
 *
 *	If row_ptr_rtn is not NULL and there exists a selected row then
 *	*row_ptr_rtn will be updated with the pointer to the row
 *	structure.
 *
 *	Returns -1 on error or no row selected, if a valid row index is
 *	returned it can be considered to be valid and allocated.
 */
gint EDVCListGetSelectedLast(
        GtkCList *clist, GtkCListRow **row_ptr_rtn
)
{
	gint i, row;
	GList *glist;


	if(row_ptr_rtn != NULL)
	    *row_ptr_rtn = NULL;

	if(clist == NULL)
	    return(-1);

	/* Get last selected row index or -1 if there is none. */
	glist = clist->selection_end;
	row = (glist != NULL) ? (gint)glist->data : -1;

	/* Selected row in bounds? */
	if((row >= 0) && (row < clist->rows))
	{
	    /* Itterate through actual row data pointers on clist. */
	    glist = clist->row_list;
	    for(i = 0; i < clist->rows; i++)
	    {
		if(glist == NULL)
		    break;

		if(i == row)
		{
		    if(row_ptr_rtn != NULL)
			*row_ptr_rtn = (GtkCListRow *)glist->data;

		    return(row);
		}

		glist = glist->next;
	    }
	}

	return(-1);
}

/*
 *      Returns the pointer of the last selected node on the given ctree.
 *
 *      If row_ptr_rtn is not NULL and there exists a selected node then
 *      *row_ptr_rtn will be updated with the pointer to the node's row
 *	structure.
 *
 *      Returns NULL on error or no node selected, if a valid node is
 *      returned it can be considered to be valid and allocated.
 */
GtkCTreeNode *EDVCTreeGetSelectedLast(
        GtkCTree *ctree, GtkCTreeRow **row_ptr_rtn
)
{
	gint i;
	GList *glist;
	GtkCList *clist;
	GtkCTreeNode *node;
	GtkCTreeRow *row_ptr;


        if(row_ptr_rtn != NULL)
            *row_ptr_rtn = NULL;

        if(ctree == NULL)
            return(NULL);

	clist = GTK_CLIST(ctree);

        /* Get last selected node or -1 if there is none. */
        glist = clist->selection_end;
        node = (glist != NULL) ? (GtkCTreeNode *)glist->data : NULL;

        /* Is there a selected node? */
        if(node != NULL)
        {
            /* Itterate through actual row data pointers on clist. */
	    row_ptr = GTK_CTREE_ROW(node);
            glist = clist->row_list;
            for(i = 0; i < clist->rows; i++)
            {
                if(glist == NULL)
                    break;

                if(row_ptr == (GtkCTreeRow *)glist->data)
                {
                    if(row_ptr_rtn != NULL)
                        *row_ptr_rtn = row_ptr;

                    return(node);
                }

                glist = glist->next;
            }
        }

	return(NULL);
}


/*
 *	Returns the column and row indexes of the given GtkCTreeNode node
 *	relative to the given GtkCTree widget.
 *
 *	Returns TRUE if the indexes are found or FALSE on failed match.
 */
gbool EDVNodeGetIndex(
	GtkCTree *ctree, GtkCTreeNode *node,
	gint *row_rtn, gint *column_rtn
)
{
	gint i;
	GList *glist;
	GtkCList *clist;
	GtkCListRow *clist_row_ptr;
	GtkCTreeRow *ctree_row_ptr;


	/* Reset returns. */
	if(row_rtn != NULL)
	    *row_rtn = -1;
	if(column_rtn != NULL)
	    *column_rtn = -1;

	if((ctree == NULL) || (node == NULL))
	    return(FALSE);

	clist = GTK_CLIST(ctree);

	/* Get clist row pointer from the given node. */
	ctree_row_ptr = GTK_CTREE_ROW(node);
	if(ctree_row_ptr == NULL)
	    return(FALSE);
	clist_row_ptr = (GtkCListRow *)ctree_row_ptr;


	/* Count rows untill we encounter the given node's row. */
	i = 0;
	glist = clist->row_list;
	while(glist != NULL)
	{
	    if(clist_row_ptr == (GtkCListRow *)glist->data)
		break;

	    i++;
	    glist = glist->next;
	}

	/* Failed to find clist row? */
	if(glist == NULL)
	    return(FALSE);

	/* Update returns. */
	if(row_rtn != NULL)
	    *row_rtn = i;
	/* Column is always 0. */
	if(column_rtn != NULL)
	    *column_rtn = 0;

	return(TRUE);
}

/*
 *	Returns the tree node that matches the given coordinates on the
 *	given GtkCTree widget.
 */
GtkCTreeNode *EDVNodeGetByCoordinates(
        GtkCTree *ctree, gint x, gint y
)
{
	gint row, column;
	GtkCList *clist;


	if(ctree == NULL)
	    return(NULL);

	clist = GTK_CLIST(ctree);

	/* Find row and column based on given coordinates. */
	if(!gtk_clist_get_selection_info(
	    clist, x, y, &row, &column
	))
	{
	    row = clist->rows;
	    column = 0;
	}
	if(row < 0)
	    row = 0;

	/* Get tree node matching the calculated row. */
	return(gtk_ctree_node_nth(ctree, row));
}

/*
 *	Scroll the given clist to the given position.
 */
void EDVScrollCListToPosition(
        GtkCList *clist, gfloat hpos, gfloat vpos
)
{
	GtkAdjustment *adj;

	if(clist == NULL)
	    return;

	adj = clist->hadjustment;
	if(adj != NULL)
	{
	    if(hpos > (adj->upper - adj->page_size))
		hpos = adj->upper - adj->page_size;
	    if(hpos < adj->lower)
		hpos = adj->lower;
	    adj->value = hpos;
	    gtk_signal_emit_by_name(GTK_OBJECT(adj), "value_changed");
	}

        adj = clist->vadjustment;
        if(adj != NULL)
        {
            if(vpos > (adj->upper - adj->page_size))
                vpos = adj->upper - adj->page_size;
            if(vpos < adj->lower)
                vpos = adj->lower;
            adj->value = vpos;
            gtk_signal_emit_by_name(GTK_OBJECT(adj), "value_changed");
        }
}


/*
 *	Returns TRUE if the given disk object name has a valid notation.
 */
gbool EDVObjectNameNotationOK(const gchar *name)
{
	gchar *strptr;


	if(name == NULL)
	    return(FALSE);

	/* No empty names allowed. */
	if(*name == '\0')
	    return(FALSE);

	/* No special notations allowed. */
        if(!strcmp(name, ".."))
            return(FALSE);
        if(!strcmp(name, "."))
            return(FALSE);

	/* No blanks allowed as first char. */
	if(*name == ' ')
            return(FALSE);
        if(*name == '\t')
            return(FALSE);

	/* No deliminators allowed. */
	strptr = strchr(name, DIR_DELIMINATOR);
	if(strptr != NULL)
	    return(FALSE);

	/* All checks passed! */
	return(TRUE);
}

/*
 *      Returns TRUE if the object is `hidden' by checking the given name
 *      of the object and testing if the first character if the name is
 *      a '.' character.
 */
gbool EDVCheckObjectHidden(
        edv_core_struct *core_ptr, edv_object_struct *object
)
{
	const gchar *name;


        if(object == NULL)
            return(FALSE);

	name = object->name;
	if(name == NULL)
	    return(FALSE);

        /* Check if the first character of the name starts with a '.'
         * character.
         */
        if(*name == '.')
	{
	    /* Special directory notations should not count in this
	     * check.
	     */
	    if(!strcmp(name, "..") || !strcmp(name, "."))
		return(FALSE);
	    else
		return(TRUE);
	}
        else
	{
            return(FALSE);
	}
}

/*
 *	Returns TRUE if the write protect is enabled on the given core
 *	structure's configuration list.
 *
 *	If verbose is set to TRUE and write protect is enabled then a 
 *	standard message will be printed to the user indicating that write
 *	protect is enabled and how to disable it.
 */
gbool EDVCheckWriteProtect(edv_core_struct *core_ptr, gbool verbose)
{
	gbool write_protect;


	if(core_ptr == NULL)
	    return(FALSE);

	/* Get write protect state from the core structure's configuration
	 * list.
	 */
	write_protect = EDVCFGItemListGetValueI(
	    core_ptr->cfg_list, EDV_CFG_PARM_WRITE_PROTECT
	);
	/* Is write protect enabled and verbose specified? */
	if(write_protect && verbose)
	{
            CDialogSetTransientFor(NULL);
            CDialogGetResponse(
#ifdef PROG_LANGUAGE_ENGLISH
"Permission Denied",
"The specified operation is not permitted because\n\
write protect is on.\n\
\n\
If you wish to perform this operation then you must turn\n\
off write protect by going to Settings->Write Protect.",
		NULL,
#endif
#ifdef PROG_LANGUAGE_SPANISH
"El Permiso Neg",
"La operacin especificada no se permite porque escribe\n\
protege est activado. Si usted desea realizar la operacin\n\
entonces usted debe apagar escribe protege yendo a\n\
Settings->Write Protect.",
                NULL,
#endif
#ifdef PROG_LANGUAGE_FRENCH
"La Permission A Ni",
"L'opration spcifie n'est pas permise parce que protge\n\
en criture est sur. Si vous souhaitez excuter l'opration\n\
alors vous devez tourner de protge en criture en allant \n\
Settings->Write Protect.",
                NULL,
#endif
                CDIALOG_ICON_WARNING,
                CDIALOG_BTNFLAG_OK,
                CDIALOG_BTNFLAG_OK
            );
            CDialogSetTransientFor(NULL);
	}

	return(write_protect);
}


/*
 *	Checks if the given effective user id owns the object who's
 *	owner is the owner_id.
 */
gbool EDVCheckUIDIsOwner(gint effective_uid, gint owner_id)
{
	return(
	   (effective_uid == 0) ||
	   (effective_uid == owner_id)
	);
}

/*
 *	Checks if the effective user id of this process owns the object
 *	sho's owner is the owner_id.
 */
gbool EDVCheckIsOwner(edv_core_struct *core_ptr, gint owner_id)
{
	if(core_ptr == NULL)
	    return(FALSE);
	else
	    return(EDVCheckUIDIsOwner(core_ptr->effective_user_id, owner_id));
}

/*
 *	Checks if the user or group id of the process can access the
 *	given object. Where access is defined as the object being
 *	executable, since any directory set x means it is accessable.
 */
gbool EDVCheckObjectAccessable(
        edv_core_struct *core_ptr, edv_object_struct *object
)
{
	gint euid, eguid;
	guint p;


	if((core_ptr == NULL) || (object == NULL))
	    return(FALSE);

	euid = core_ptr->effective_user_id;
        eguid = core_ptr->effective_group_id;
	p = object->permissions;

	/* Sticky and atleast one catagory has execute permission? */
	if((p & EDV_PERMISSION_STICKY) &&
	   (p & (EDV_PERMISSION_UEXECUTE | EDV_PERMISSION_GEXECUTE |
                 EDV_PERMISSION_AEXECUTE))
	)
	    return(TRUE);	

	/* Anyone can execute? */
	if(p & EDV_PERMISSION_AEXECUTE)
	    return(TRUE);

	/* Group can execute and is member of group? */
	if((p & EDV_PERMISSION_GEXECUTE) &&
	   (eguid <= object->group_id)
	)
	    return(TRUE);

	/* Owner can execute and is a owner? */
        if((p & EDV_PERMISSION_UEXECUTE) &&
           ((euid == 0) || (euid == object->owner_id))
        )
            return(TRUE);

	return(FALSE);
}

/*
 *      Checks if the user or group id of the process can read the
 *      given object.
 */
gbool EDVCheckObjectReadable(
        edv_core_struct *core_ptr, edv_object_struct *object
)
{
        gint euid, eguid;
        guint p;


        if((core_ptr == NULL) || (object == NULL))
            return(FALSE);

        euid = core_ptr->effective_user_id;
        eguid = core_ptr->effective_group_id;
        p = object->permissions;

        /* Sticky and atleast one catagory has read permission? */
        if((p & EDV_PERMISSION_STICKY) &&
           (p & (EDV_PERMISSION_UREAD | EDV_PERMISSION_GREAD |
                 EDV_PERMISSION_AREAD))
        )
            return(TRUE);

        /* Anyone can read? */
        if(p & EDV_PERMISSION_AREAD)
            return(TRUE);

        /* Group can read and is member of group? */
        if((p & EDV_PERMISSION_GREAD) &&
           (eguid <= object->group_id)
        )
            return(TRUE);

        /* Owner can read and is a owner? */
        if((p & EDV_PERMISSION_UREAD) &&
           ((euid == 0) || (euid == object->owner_id))
        )
            return(TRUE);

        return(FALSE);
}

/*
 *      Checks if the user or group id of the process can write the
 *      given object.
 */
gbool EDVCheckObjectWriteable(
        edv_core_struct *core_ptr, edv_object_struct *object
)
{
        gint euid, eguid;
        guint p;


        if((core_ptr == NULL) || (object == NULL))
            return(FALSE);

        euid = core_ptr->effective_user_id;
        eguid = core_ptr->effective_group_id;
        p = object->permissions;

        /* Sticky and atleast one catagory has write permission? */
        if((p & EDV_PERMISSION_STICKY) &&
           (p & (EDV_PERMISSION_UWRITE | EDV_PERMISSION_GWRITE |
                 EDV_PERMISSION_AWRITE))
        )
            return(TRUE);

        /* Anyone can write? */
        if(p & EDV_PERMISSION_AWRITE)
            return(TRUE);

        /* Group can write and is member of group? */
        if((p & EDV_PERMISSION_GWRITE) &&
           (eguid <= object->group_id)
        )
            return(TRUE);

        /* Owner can write and is a owner? */
        if((p & EDV_PERMISSION_UWRITE) &&
           ((euid == 0) || (euid == object->owner_id))
        )
            return(TRUE);

        return(FALSE);
}

/*
 *      Checks if the user or group id of the process can execute the
 *      given object.
 */
gbool EDVCheckObjectExecutable(
        edv_core_struct *core_ptr, edv_object_struct *object
)
{
        gint euid, eguid;
        guint p;


        if((core_ptr == NULL) || (object == NULL))
            return(FALSE);

        euid = core_ptr->effective_user_id;
        eguid = core_ptr->effective_group_id;
        p = object->permissions;

        /* Sticky and atleast one catagory has execute permission? */
        if((p & EDV_PERMISSION_STICKY) &&
           (p & (EDV_PERMISSION_UEXECUTE | EDV_PERMISSION_GEXECUTE |
                 EDV_PERMISSION_AEXECUTE))
        )
            return(TRUE);

        /* Anyone can execute? */
        if(p & EDV_PERMISSION_AEXECUTE)
            return(TRUE);

        /* Group can execute and is member of group? */
        if((p & EDV_PERMISSION_GEXECUTE) &&
           (eguid <= object->group_id)
        )
            return(TRUE);

        /* Owner can execute and is a owner? */
        if((p & EDV_PERMISSION_UEXECUTE) &&
           ((euid == 0) || (euid == object->owner_id))
        )
            return(TRUE);

        return(FALSE);
}

/*
 *	Returns TRUE if the object specified by path is supported by
 *	Imlib.
 */
gbool EDVCheckImlibImage(
        edv_core_struct *core_ptr, const gchar *path
)
{
        gint i;
        const gchar *ext_list[] = IMLIB_IMAGE_FILE_EXTENSIONS;


	if((core_ptr == NULL) || (path == NULL))
	    return(FALSE);

        /* Check if this is an image file by checking the extension
         * portion of the name.
         */
        for(i = 0; ext_list[i] != NULL; i += 2)
        {
	    if(EDVIsExtension(path, ext_list[i]))
                return(TRUE);
        }

	return(FALSE);
}

/*
 *      Returns TRUE if the object specified by path is an archive
 *      supported by Endeavour's archiver.
 */
gbool EDVCheckEDVArchiverArchive(
        edv_core_struct *core_ptr, const gchar *path
)
{
        gint i;
        const gchar *ext_list[] = EDV_ARCHIVER_ARCHIVE_FILE_EXTENSIONS;


        if((core_ptr == NULL) || (path == NULL))
            return(FALSE);

        /* Check if this is an image file by checking the extension
         * portion of the name.
         */
        for(i = 0; ext_list[i] != NULL; i += 2)
        {
            if(EDVIsExtension(path, ext_list[i]))
                return(TRUE);
        }

        return(FALSE);
}


/*
 *	Fetches the pixmap and mask pairs that best matches the given
 *	information. Devices and MIME Types will be realized as needed
 *	during this call.
 *
 *	Any of the given information may be NULL or 0.
 *
 *	The returned pixmap and mask pairs do not have any refcounts
 *	added to them.
 *
 *	Returns 0 on success or -1 on failed match.
 */
gint EDVMatchObjectIcon(
        edv_device_struct **device, gint total_devices,
        edv_mimetype_struct **mimetype, gint total_mimetypes,
        gint type,              /* One of EDV_OBJECT_TYPE_*. */
        const gchar *path,      /* Full path or just name. */
        gbool link_valid, guint permissions,
	gint icon_size,		/* 0 = small, 1 = medium, 2 = large. */
        GdkPixmap **pixmap_closed, GdkBitmap **mask_closed,
        GdkPixmap **pixmap_opened, GdkBitmap **mask_opened,
        GdkPixmap **pixmap_extended, GdkBitmap **mask_extended
)
{
	gbool got_match = FALSE;
	gint i;
	const gchar *cstrptr, *ext = NULL;
	edv_device_struct *dev_ptr;
	edv_mimetype_struct *mt_ptr;
	GdkPixmap *pixmap[3];
	GdkBitmap *mask[3];


	/* Reset locals. */
	memset(pixmap, 0x00, 3 * sizeof(GdkPixmap *));
        memset(mask, 0x00, 3 * sizeof(GdkBitmap *));

	/* Reset returns. */
	if(pixmap_closed != NULL)
	    *pixmap_closed = NULL;
        if(mask_closed != NULL)
            *mask_closed = NULL;

        if(pixmap_opened != NULL)
            *pixmap_opened = NULL;
        if(mask_opened != NULL)
            *mask_opened = NULL;

        if(pixmap_extended != NULL)
            *pixmap_extended = NULL;
        if(mask_extended != NULL)
            *mask_extended = NULL;


	/* Parse extension from the given path if it is not NULL. */
	if(path != NULL)
	{
	    /* Seek cstrptr to child portion of path. */
            cstrptr = strrchr(path, DIR_DELIMINATOR);
            if(cstrptr != NULL)
                cstrptr++;
            else
                cstrptr = path;

	    /* Child name starts with a '.'? */
            if(*cstrptr == '.')
                cstrptr++;

	    /* Seek to extension (if any), include the '.' as the first
	     * character if an extension is found.
	     */
            ext = strchr(cstrptr, '.');
	    if(ext != NULL)
	    {
		/* Extension must have atleast one character after it. */
		if(*(ext + 1) == '\0')
		    ext = NULL;
	    }
	}

/* Macro to set the pixmap and mask array values to the MIME Type
 * specified by mt_ptr and realize the MIME Type. If one or more pixmaps
 * were obtained than got_match will be set to TRUE.
 */
#define DO_LOAD_MT_ICONS        \
{ \
 gint n; \
 gint min = MIN(EDV_MIMETYPE_TOTAL_ICON_STATES, 3); \
\
\
 /* Realize MIME Type first (as needed) to ensure icons are loaded. */ \
 EDVMimeTypeRealize(mt_ptr, FALSE); \
\
 /* Get icons by the requested size. */ \
 switch(icon_size) \
 { \
  case 0: \
   for(n = 0; n < min; n++) \
   { \
    pixmap[n] = mt_ptr->small_pixmap[n]; \
    mask[n] = mt_ptr->small_mask[n]; \
   } \
   break; \
\
  case 1: \
   for(n = 0; n < min; n++) \
   { \
    pixmap[n] = mt_ptr->medium_pixmap[n]; \
    mask[n] = mt_ptr->medium_mask[n]; \
   } \
   break; \
\
  case 2: \
   for(n = 0; n < min; n++) \
   { \
    pixmap[n] = mt_ptr->large_pixmap[n]; \
    mask[n] = mt_ptr->large_mask[n]; \
   } \
   break; \
 } \
\
 /* Check if we got a valid pixmap, if we did then set got_match to \
  * TRUE. \
  */ \
 for(n = 0; n < min; n++) \
 { \
  if(pixmap[n] != NULL) \
  { \
   got_match = TRUE; \
   break; \
  } \
 } \
}


	/* First check if path is a link. */
	if(type == EDV_OBJECT_TYPE_LINK)
	{
	    const gchar *mime_type_str = EDV_MIMETYPE_STR_LINK;

            for(i = 0; i < total_mimetypes; i++)
            {
                mt_ptr = mimetype[i];
                if(mt_ptr == NULL)
                    continue;

                /* Only handle if MIME Type class is a systems object. */
                if((mt_ptr->mt_class == EDV_MIMETYPE_CLASS_SYSTEM) &&
                   (mt_ptr->type != NULL)
                )
                {
                    if(!strcmp(mt_ptr->type, mime_type_str))
                    {
                        DO_LOAD_MT_ICONS
                        got_match = TRUE;
                        break;
                    }
                }
            }
	}


	/* Check devices if device list is given and path is an
	 * absolute path.
	 */
	if((device != NULL) &&
           ((path != NULL) ? ISPATHABSOLUTE(path) : FALSE) &&
	   !got_match
	)
	{
	    for(i = 0; i < total_devices; i++)
	    {
		dev_ptr = device[i];
		if(dev_ptr == NULL)
		    continue;

		if(dev_ptr->mount_path == NULL)
		    continue;

		if(!strcmp(dev_ptr->mount_path, path))
		{
		    gint n;
		    gint min = MIN(EDV_DEVICE_TOTAL_ICON_STATES, 3);


		    /* Realize this device first to ensure that the icon
		     * images are loaded.
		     */
		    EDVDeviceRealize(dev_ptr, FALSE);

		    /* Got full path match for device's mounted path. */
		    switch(icon_size)
		    {
		      case 0:
			for(n = 0; n < min; n++)
			{
			    pixmap[n] = dev_ptr->small_pixmap[n];
                            mask[n] = dev_ptr->small_mask[n];
			}
			break;

                      case 1:
                        for(n = 0; n < min; n++)
                        {
                            pixmap[n] = dev_ptr->medium_pixmap[n];
                            mask[n] = dev_ptr->medium_mask[n];
                        }
                        break;

                      case 2:
                        for(n = 0; n < min; n++)
                        {
                            pixmap[n] = dev_ptr->large_pixmap[n];
                            mask[n] = dev_ptr->large_mask[n];
                        }
                        break;
		    }
		    /* Stop itterating through devices once we got a
		     * match. Mark got match only if an actual valid pixmap
		     * was obtained.
		     */
		    for(n = 0; n < min; n++)
		    {
			if(pixmap[n] != NULL)
			{
			    got_match = TRUE;
			    break;
			}
		    }
		    break;
		}
	    }
	}

        /* Check MIME Types. */
        if((mimetype != NULL) && !got_match)
	{
	    const gchar *value;


            for(i = 0; i < total_mimetypes; i++)
            {
                mt_ptr = mimetype[i];
                if(mt_ptr == NULL)
                    continue;

		value = mt_ptr->value;
		if(value == NULL)
		    continue;


		/* Handle by MIME Type class. */
		switch(mt_ptr->mt_class)
		{
		  case EDV_MIMETYPE_CLASS_UNIQUE:
                  case EDV_MIMETYPE_CLASS_PROGRAM:
		    if((path != NULL) ? ISPATHABSOLUTE(path) : FALSE)
		    {
			if(!strcmp(value, path))
			    DO_LOAD_MT_ICONS
		    }
		    break;

		  case EDV_MIMETYPE_CLASS_FORMAT:
                    if((path != NULL) && !got_match)
                    {
			if(EDVIsExtension(path, value))
                            DO_LOAD_MT_ICONS
                    }
                    break;

		}
		/* Do not stop itterating through MIME Types if got_match
		 * is TRUE, because there might be another MIME type that
		 * has a more specific icon further in the list.
		 */
	    }
	}



	/* If still did not get match then use basic system MIME types
	 * for the given object.
	 */
	if(!got_match)
	{
	    const gchar *mime_type_str;


	    /* Get MIME Type type string used for matching of system
	     * object type.
	     */
	    switch(type)
	    {
	      case EDV_OBJECT_TYPE_FILE:
		/* Check the file's permissions allow execution, in which
		 * case we use the file/executable MIME Type instead of
		 * file/regular.
		 */
		if(permissions & (EDV_PERMISSION_UEXECUTE |
		    EDV_PERMISSION_GEXECUTE | EDV_PERMISSION_AEXECUTE)
		)
		    mime_type_str = EDV_MIMETYPE_STR_FILE_EXECUTABLE;
		else
		    mime_type_str = EDV_MIMETYPE_STR_FILE;
		break;

	      case EDV_OBJECT_TYPE_DIRECTORY:
		mime_type_str = EDV_MIMETYPE_STR_DIRECTORY;
                break;

              case EDV_OBJECT_TYPE_LINK:
                mime_type_str = EDV_MIMETYPE_STR_LINK;
                break;

              case EDV_OBJECT_TYPE_DEVICE_BLOCK:
                mime_type_str = EDV_MIMETYPE_STR_DEVICE_BLOCK;
                break;

              case EDV_OBJECT_TYPE_DEVICE_CHARACTER:
                mime_type_str = EDV_MIMETYPE_STR_DEVICE_CHARACTER;
                break;

              case EDV_OBJECT_TYPE_FIFO:
                mime_type_str = EDV_MIMETYPE_STR_FIFO;
                break;

              case EDV_OBJECT_TYPE_SOCKET:
                mime_type_str = EDV_MIMETYPE_STR_SOCKET;
                break;

	      default:
		mime_type_str = EDV_MIMETYPE_STR_FILE;
		break;
            }

	    /* Iterate through MIME Types of class system. */
            for(i = 0; i < total_mimetypes; i++)
            {
                mt_ptr = mimetype[i];
                if(mt_ptr == NULL)
                    continue;

                /* Only handle if MIME Type class is a systems object. */
                if((mt_ptr->mt_class == EDV_MIMETYPE_CLASS_SYSTEM) &&
                   (mt_ptr->type != NULL)
		)
                {
		    if(!strcmp(mt_ptr->type, mime_type_str))
		    {
			DO_LOAD_MT_ICONS
			got_match = TRUE;
			break;
		    }
		}
	    }

	    /* If we still didn't get match, then all else return the
	     * file icons.
	     */
/* Work on this later. */
	}
#undef DO_LOAD_MT_ICONS


	/* Update returns. */
        if(pixmap_closed != NULL)
            *pixmap_closed = pixmap[0];
        if(mask_closed != NULL)
            *mask_closed = mask[0];

        if(pixmap_opened != NULL)
            *pixmap_opened = pixmap[1];
        if(mask_opened != NULL)
            *mask_opened = mask[1];

        if(pixmap_extended != NULL)
            *pixmap_extended = pixmap[2];
        if(mask_extended != NULL)
            *mask_extended = mask[2];

	return(0);
}

/*
 *      Fetches the MIME Type type string that best matches the given
 *      information.
 *
 *      Any of the given information may be NULL or 0.
 *
 *      The returned MIME Type string must not be deallocated.
 *
 *      Returns 0 on success or -1 on failed match.
 */
gint EDVMatchObjectTypeString(
        edv_mimetype_struct **mimetype, gint total_mimetypes,
        gint type,              /* One of EDV_OBJECT_TYPE_*. */
	guint permissions,
        const gchar *path,      /* Full path or just name. */
        gchar **type_str
)
{
        gbool got_match = FALSE;
        gint i;
	gchar *ltype_str = NULL;
        const gchar *cstrptr, *ext = NULL;
        edv_mimetype_struct *mt_ptr;


	if(type_str != NULL)
	    *type_str = NULL;


        /* Parse extension from the given path if it is not NULL. */
        if(path != NULL)
        {
            /* Seek cstrptr to child portion of path. */
            cstrptr = strrchr(path, DIR_DELIMINATOR);
            if(cstrptr != NULL)
                cstrptr++;
            else
                cstrptr = path;

            /* Child name starts with a '.'? */
            if(*cstrptr == '.')
                cstrptr++;

            /* Seek to extension (if any), include the '.' as the first
             * character if an extension is found.
             */
            ext = strchr(cstrptr, '.');
            if(ext != NULL)
            {
                /* Extension must have atleast one character after it. */
                if(*(ext + 1) == '\0')
                    ext = NULL;
            }
        }


        /* Check MIME Types first. */
        if(mimetype != NULL)
        {
            const gchar *value;


            for(i = 0; i < total_mimetypes; i++)
            {
                mt_ptr = mimetype[i];
                if(mt_ptr == NULL)
                    continue;

                value = mt_ptr->value;
                if(value == NULL)
                    continue;

                /* Handle by MIME Type class. */
                switch(mt_ptr->mt_class)
                {
                  case EDV_MIMETYPE_CLASS_UNIQUE:
                  case EDV_MIMETYPE_CLASS_PROGRAM:
                    if((path != NULL) ? ISPATHABSOLUTE(path) : FALSE)
                    {
                        if(!strcmp(value, path))
                        {
                            ltype_str = mt_ptr->type;
			    if(ltype_str != NULL)
				got_match = TRUE;
                        }
                    }
                    break;

                  case EDV_MIMETYPE_CLASS_FORMAT:
                    if((path != NULL) && !got_match)
                    {
                        if(EDVIsExtension(path, value))
                        {
                            ltype_str = mt_ptr->type;
			    if(ltype_str != NULL)
				got_match = TRUE;
                        }
		    }
		    break;
                }
                /* Do not stop itterating through MIME Types if got_match
                 * is TRUE, because there might be another MIME type that
                 * has a more specific icon further in the list.
                 */
            }
        }

        /* If still did not get match then use basic system MIME types
         * for the given object.
         */
        if(!got_match)
        {
            const gchar *mime_type_str;


            /* Get MIME Type type string used for matching of system
             * object type.
             */
            switch(type)
            {
              case EDV_OBJECT_TYPE_FILE:
                /* Check the file's permissions allow execution, in which
                 * case we use the file/executable MIME Type instead of
                 * file/regular.
                 */
                if(permissions & (EDV_PERMISSION_UEXECUTE |
                    EDV_PERMISSION_GEXECUTE | EDV_PERMISSION_AEXECUTE)
                )
                    mime_type_str = EDV_MIMETYPE_STR_FILE_EXECUTABLE;
		else
		    mime_type_str = EDV_MIMETYPE_STR_FILE;
                break;

              case EDV_OBJECT_TYPE_DIRECTORY:
                mime_type_str = EDV_MIMETYPE_STR_DIRECTORY;
                break;

              case EDV_OBJECT_TYPE_LINK:
                mime_type_str = EDV_MIMETYPE_STR_LINK;
                break;

              case EDV_OBJECT_TYPE_DEVICE_BLOCK:
                mime_type_str = EDV_MIMETYPE_STR_DEVICE_BLOCK;
                break;

              case EDV_OBJECT_TYPE_DEVICE_CHARACTER:
                mime_type_str = EDV_MIMETYPE_STR_DEVICE_CHARACTER;
                break;

              case EDV_OBJECT_TYPE_FIFO:
                mime_type_str = EDV_MIMETYPE_STR_FIFO;
                break;

              case EDV_OBJECT_TYPE_SOCKET:
                mime_type_str = EDV_MIMETYPE_STR_SOCKET;
                break;

              default:
                mime_type_str = EDV_MIMETYPE_STR_FILE;
                break;
            }

	    ltype_str = (gchar *)mime_type_str;
	}


	/* Update returns. */
        if(type_str != NULL)
            *type_str = ltype_str;


	return(0);
}


/*
 *	Opens the file specified by path and displays its contents on
 *	the cdialog.
 */
void EDVShowFileInCDialog(
        const gchar *path, const gchar *title, gint cdialog_icon,
        GtkWidget *toplevel
)
{
	FILE *fp;
	gchar *buf;
	gint buf_len, bytes_read;
	struct stat stat_buf;



	if(path == NULL)
	    return;

	/* Open file for reading. */
	fp = FOpen(path, "rb");
	if(fp == NULL)
	    return;

	/* Get statistics of opened file. */
	if(fstat(fileno(fp), &stat_buf))
	{
	    FClose(fp);
            return;
	}

	/* Get length of buffer and allocate read buffer to match the size
	 * of the file up to 10000 bytes plus one extra byte for the null
	 * terminator byte.
	 */
	buf_len = (gint)((stat_buf.st_size > 10000) ?
	    10000 : stat_buf.st_size
	);
	if(buf_len < 0)
	    buf_len = 0;
	buf = (gchar *)g_malloc((buf_len + 1) * sizeof(gchar));
	if(buf == NULL)
	{
	    FClose(fp);
	    return;
	}

	/* Read file. */
	bytes_read = (gint)fread(buf, sizeof(gchar), buf_len, fp);

	/* Close file. */
	FClose(fp);
	fp = NULL;

	/* Error reading file? */
	if(bytes_read < 0)
	{
	    g_free(buf);
	    return;
	}

	/* Null terminate. */
	if(bytes_read <= buf_len)
	    buf[bytes_read] = '\0';
	else
	    buf[buf_len] = '\0';

	/* File contents empty? */
	if(*buf == '\0')
	{
	    g_free(buf);
	    return;
	}

	/* Display read file contents on cdialog. */
	CDialogSetTransientFor(toplevel);
	CDialogGetResponse(
	    title, buf, NULL,
	    cdialog_icon,
	    CDIALOG_BTNFLAG_OK,
	    CDIALOG_BTNFLAG_OK
	);
	CDialogSetTransientFor(NULL);

	/* Deallocate read file contents buffer. */
	g_free(buf);
	buf = NULL;
}


/*
 *	Sets up the given widget (which must be a GtkEntry) to allow
 *	it to receive dragged disk object paths.
 *
 *	Only the text on the GtkEntry will be updated when a disk object
 *	path is dragged on to it, no "activate" signal will be emitted.
 */
void EDVEntrySetDND(edv_core_struct *core_ptr, GtkWidget *w)
{
        const GtkTargetEntry dnd_tar_types[] = {
{"text/plain",                          0,      EDV_DND_TYPE_INFO_TEXT_PLAIN},
{"text/uri-list",                       0,      EDV_DND_TYPE_INFO_TEXT_URI_LIST},
{"STRING",                              0,      EDV_DND_TYPE_INFO_STRING}
        };

	if((core_ptr == NULL) || (w == NULL))
	    return;

	if(!GTK_IS_ENTRY(w))
	    return;

        GUIDNDSetTar(
            w,
            dnd_tar_types,
            sizeof(dnd_tar_types) / sizeof(GtkTargetEntry),
            GDK_ACTION_COPY,			/* Actions. */
            GDK_ACTION_COPY,			/* Default action if same. */
            GDK_ACTION_COPY,			/* Default action. */
            EDVEntryDragDataReceivedCB,
            core_ptr
        );
        gtk_signal_connect_after(
            GTK_OBJECT(w), "drag_motion",
            GTK_SIGNAL_FUNC(EDVEntryDragMotionCB), core_ptr
        );
}

/*
 *	Appends a new operation history event item to the list on the
 *	given core structure.
 */
void EDVAppendHistory(
        edv_core_struct *core_ptr,
        gint type,              /* One of EDV_HISTORY_*. */
        gulong time_start,      /* Time the operation first started. */
        gulong time_end,        /* Time the operation ended. */
        gint status,
        const gchar *src_path, const gchar *tar_path,
        const gchar *comments
)
{
	gint max;
	edv_history_struct *h;


	if(core_ptr == NULL)
	    return;

	/* Get maximum history event items from configuration list on the
	 * core structure.
	 */
	max = EDVCFGItemListGetValueI(
	    core_ptr->cfg_list, EDV_CFG_PARM_HISTORY_EVENTS_MAX
	);

	/* Append a new history event item and discard any old items. */
	h = EDVHistoryListAppendNew(
	    &core_ptr->history_event, &core_ptr->total_history_events,
	    NULL, max
	);
	if(h != NULL)
	{
	    EDVHistoryDiskObjectSet(
		h, type, time_start, time_end, status,
		src_path, tar_path, comments
	    );
	}
}
