#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include "../include/string.h"
#include "../include/disk.h"

#include "guiutils.h"
#include "toolbar.h"

#include "edvtypes.h"
#include "edvcfg.h"
#include "edvobj.h"
#include "edvdevices.h"
#include "edvstatusbar.h"
#include "fopdialog.h"
#include "imbr.h"
#include "imbrcb.h"
#include "imbropcb.h"
#include "imbrtlist.h"
#include "imbrdnd.h"
#include "endeavour.h"
#include "edvcb.h"
#include "edvutils.h"
#include "edvcfglist.h"
#include "config.h"

#include "images/icon_save_20x20.xpm"
#include "images/icon_close_20x20.xpm"
#include "images/icon_exit_20x20.xpm"
#include "images/icon_move_file_20x20.xpm"
#include "images/icon_copy_file_20x20.xpm"
#include "images/icon_rename_20x20.xpm"
#include "images/icon_cancel_20x20.xpm"
#include "images/icon_chmod_20x20.xpm"
#include "images/icon_owned_20x20.xpm"
#include "images/icon_properties2_20x20.xpm"
#include "images/icon_stop_20x20.xpm"
#include "images/icon_run_20x20.xpm"
#include "images/icon_reload_20x20.xpm"
#include "images/icon_search_20x20.xpm"
#include "images/icon_device_misc_20x20.xpm"
#include "images/icon_tool_bar_20x20.xpm"
#include "images/icon_status_bar_20x20.xpm"
#include "images/icon_options2_20x20.xpm"
#include "images/icon_customize_20x20.xpm"
#include "images/icon_insecure_20x20.xpm"
#include "images/icon_secure_20x20.xpm"

#include "images/icon_open_20x20.xpm"
#include "images/icon_file_20x20.xpm"
#include "images/icon_folder_closed_20x20.xpm"
#include "images/icon_folder_parent_20x20.xpm"
#include "images/icon_link2_20x20.xpm"
#include "images/icon_pipe_20x20.xpm"
#include "images/icon_device_block_20x20.xpm"
#include "images/icon_device_character_20x20.xpm"
#include "images/icon_socket_20x20.xpm"

#include "images/icon_trash_empty_20x20.xpm"
#include "images/icon_trash_20x20.xpm"
#include "images/icon_help_20x20.xpm"
#include "images/icon_about_20x20.xpm"

#include "images/endeavour_20x20.xpm"
#include "images/image_browser_20x20.xpm"
#include "images/archiver_20x20.xpm"
#include "images/image_browser_48x48.xpm"


static edv_imbr_opid_struct *EDVImbrMatchOPID(
	edv_imbr_struct *imbr, gint op
);

void EDVImbrSetTitle(edv_imbr_struct *imbr, const gchar *path);

gchar *EDVImbrCurrentLocation(edv_imbr_struct *imbr);
void EDVImbrSetLocation(
	edv_imbr_struct *imbr,
	const gchar *path, gbool record_history
);
void EDVImbrUpdateLocationIcon(
        edv_imbr_struct *imbr, const gchar *path
);

void EDVImbrToolbarRegenerate(edv_imbr_struct *imbr);
void EDVImbrTListMenuRegenerate(edv_imbr_struct *imbr);

void EDVImbrSyncConfiguration(edv_imbr_struct *imbr);

void EDVImbrQueueLoadingProcess(edv_imbr_struct *imbr);

void EDVImbrSelectPath(
        edv_imbr_struct *imbr, const gchar *path
);

edv_imbr_opid_struct **EDVImbrOPIDListNew(
	edv_imbr_struct *imbr,		/* Can be NULL. */
	gint *total
);
void EDVImbrOPIDListDelete(
	edv_imbr_opid_struct **list, gint total
);

static void EDVImbrCreateMenuBar(
	edv_imbr_struct *imbr, GtkWidget *parent
);
edv_imbr_struct *EDVImbrNew(gpointer core_ptr);
void EDVImbrReset(edv_imbr_struct *imbr, gbool need_unmap);
void EDVImbrSyncData(edv_imbr_struct *imbr);
void EDVImbrUpdateMenus(edv_imbr_struct *imbr);
void EDVImbrSetBusy(edv_imbr_struct *imbr, gbool is_busy);
void EDVImbrMap(edv_imbr_struct *imbr);
void EDVImbrUnmap(edv_imbr_struct *imbr);
void EDVImbrDelete(edv_imbr_struct *imbr);



/*
 *	Returns the operation id structure on the given image browser that
 *	matches the given id.
 */
static edv_imbr_opid_struct *EDVImbrMatchOPID(
        edv_imbr_struct *imbr, gint op
)
{
	gint i;
	edv_imbr_opid_struct *opid;


	if(imbr == NULL)
	    return(NULL);

	for(i = 0; i < imbr->total_opids; i++)
	{
	    opid = imbr->opid[i];
	    if((opid != NULL) ? (opid->op == op) : FALSE)
		return(opid);
	}

	return(NULL);
}


/*
 *	Updates the WM title on the given image browser's toplevel window.
 *
 *	The given path is optional, it can be either the current
 *	selected location or the last selected object.
 */
void EDVImbrSetTitle(edv_imbr_struct *imbr, const gchar *path)
{
	gchar *title;
	GtkWidget *w;


	if(imbr == NULL)
	    return;

	if(path == NULL)
	    path = "";

	w = imbr->toplevel;
	if(w == NULL)
	    return;

	/* Format title string. */
	title = g_strdup_printf(
	    "Image Browser %s",
            (path != NULL) ? path : ""
	);

	/* Set new title. */
	gtk_window_set_title(GTK_WINDOW(w), title);

	/* Deallocate title string, it is no longer needed. */
	g_free(title);
}

/*
 *	Returns the path to the current location of the image browser.
 *
 *	Can return NULL on error.
 */
gchar *EDVImbrCurrentLocation(edv_imbr_struct *imbr)
{
	GtkCombo *combo;


	if(imbr == NULL)
	    return(NULL);

	combo = (GtkCombo *)imbr->location_combo;
	if(combo == NULL)
	    return(NULL);

	return(gtk_entry_get_text(GTK_ENTRY(combo->entry)));
}

/*
 *	Sets the location combo of the given image browser to the value of
 *	path.
 *
 *	If record_history is TRUE then the current value will be recorded
 *	on the combo's list before setting of the new value.
 *	Duplicate values will not be recorded.
 */
void EDVImbrSetLocation(
        edv_imbr_struct *imbr,
        const gchar *path, gbool record_history
)
{
        GtkCombo *combo;
	gchar *new_path;


        if((imbr == NULL) || (path == NULL))
            return;

        combo = (GtkCombo *)imbr->location_combo;
        if(combo == NULL)
            return;

        /* Copy given path as the new path. */
        new_path = g_strdup(path);

	/* Simplify new path, reducing occurances of /.. */
	SimplifyPath(new_path);


	/* Check for no change in value. */
	if(new_path != NULL)
	{
	    const gchar *cstrptr = gtk_entry_get_text(GTK_ENTRY(combo->entry));
	    if(cstrptr != NULL)
	    {
		/* No change in value? */
		if(!strcmp(cstrptr, new_path))
		{
		    g_free(new_path);
		    new_path = NULL;
		    return;
		}
	    }
	}

	/* Get old value from combo's entry and record it on the list. */
	if(record_history)
	{
	    gchar *old_path;
	    const gchar *cstrptr = gtk_entry_get_text(GTK_ENTRY(combo->entry));
	    old_path = (cstrptr != NULL) ? g_strdup(cstrptr) : NULL;
	    if(old_path != NULL)
	    {
		GUIComboAddItem(combo, old_path);
		g_free(old_path);
		old_path = NULL;
	    }
	}

	/* Set new value on entry and deallocate coppied new_path. */
	if(new_path != NULL)
	{
	    gtk_entry_set_text(GTK_ENTRY(combo->entry), new_path);
	    g_free(new_path);
	    new_path = NULL;
	}
}

/*
 *	Updates the location bar's icon pixmap with respect to the path
 *	specified by path.
 *
 *	If the given path is NULL then the icon will be set as unknown.
 */
void EDVImbrUpdateLocationIcon(
        edv_imbr_struct *imbr, const gchar *path
)
{
	gint type, width, height;
	guint permissions;
	GdkPixmap *pixmap_closed, *pixmap_opened;
	GdkBitmap *mask_closed, *mask_opened;
        GtkWidget *parent, *w;
	edv_core_struct *core_ptr;


	if(imbr == NULL)
	    return;

	core_ptr = (edv_core_struct *)imbr->core_ptr;
	if(core_ptr == NULL)
	    return;

	/* Get icon fixed widget as the parent. */
	parent = imbr->location_icon_fixed;
	if(parent == NULL)
	    return;


	/* Get type and permissions of object that the given path reffers
	 * to.
	 */
	type = EDV_OBJECT_TYPE_DIRECTORY;
	permissions = 0x00000000;
	if(path != NULL)
	{
	    struct stat lstat_buf;

	    /* Try to get local stats of the given path. */
	    if(!lstat(path, &lstat_buf))
	    {
		type = EDVObjectGetTypeFromStatMode(lstat_buf.st_mode);
		permissions = EDVObjectGetPermissionsFromStatMode(lstat_buf.st_mode);
	    }
	}

	/* Get pixmap and mask for icon. */
	EDVMatchObjectIcon(
	    core_ptr->device, core_ptr->total_devices,
	    core_ptr->mimetype, core_ptr->total_mimetypes,
	    type,
	    path,
	    TRUE, permissions,
	    0,			/* Small icons. */
	    &pixmap_closed, &mask_closed,
	    &pixmap_opened, &mask_opened,
	    NULL, NULL
	);
	/* If `opened' pixmap is not available then revert to `closed'. */
	if(pixmap_opened == NULL)
	{
	    pixmap_opened = pixmap_closed;
	    mask_opened = mask_closed;
	}

	/* If no `opened' version of pixmap available then give up. */
	if(pixmap_opened == NULL)
	    return;


	/* Create new icon pixmap widget as needed. */
	w = imbr->location_icon_pm;
	if(w == NULL)
	{
	    imbr->location_icon_pm = w = gtk_pixmap_new(
		pixmap_opened, mask_opened
	    );
	    gtk_fixed_put(GTK_FIXED(parent), w, 0, 0);
	    gtk_widget_show(w);
	}
	else
	{
	    gtk_pixmap_set(GTK_PIXMAP(w), pixmap_opened, mask_opened);
	}

        gdk_window_get_size(
	    (GdkWindow *)pixmap_opened, &width, &height
	);

        /* Adjust size of fixed widget to fit pixmap. */
/*
        gtk_widget_set_usize(parent, width, height);
	gtk_widget_queue_resize(parent);
 */

        /* Set new shape of fixed widget parent. */
        gtk_widget_shape_combine_mask(parent, mask_opened, 0, 0);
}


/*
 *	Recreates the tool bar on the given image browser.
 */
void EDVImbrToolbarRegenerate(edv_imbr_struct *imbr)
{
	gint i;
        GtkWidget *w, *parent;
        edv_core_struct *core_ptr;
	edv_intlist_struct *intlist;
	edv_imbr_opid_struct *opid;
	gint tb_display, tb_relief;
	toolbar_item_struct **item, *item_ptr;
	gint total_items;


        if(imbr == NULL)
            return;

        core_ptr = (edv_core_struct *)imbr->core_ptr;
        if(core_ptr == NULL)
            return;

	/* Get image browser tool bar display and relief settings. */
	tb_display = EDVCFGItemListGetValueI(
	    core_ptr->cfg_list, EDV_CFG_PARM_IMBR_TOOL_BAR_DISPLAY
	);
	tb_relief = EDVCFGItemListGetValueI(
            core_ptr->cfg_list, EDV_CFG_PARM_IMBR_TOOL_BAR_RELIEF
        );


	/* Get parent for tool bar. */
	parent = imbr->tool_bar_handle;
	if(parent == NULL)
	    return;


        /* Delete old tool bar (if any). */
        ToolBarDelete(imbr->toolbar);
	imbr->toolbar = NULL;


	/* Get tool bar arrangement from configuration list. */
	intlist = EDVCFGItemListGetValueIntList(
	    core_ptr->cfg_list, EDV_CFG_PARM_IMBR_TOOL_BAR
	);
	if(intlist == NULL)
	    return;

	/* Create tool bar items list. */
	total_items = intlist->total;
	if(total_items > 0)
	{
	    item = (toolbar_item_struct **)g_malloc0(
		total_items * sizeof(toolbar_item_struct *)
	    );
	    if(item == NULL)
		total_items = 0;
	}
	else
	{
	    item = NULL;
	}
	for(i = 0; i < total_items; i++)
	{
	    /* Get opid structure that matches the operation id specified
	     * on the intlist.
	     */
	    opid = EDVImbrMatchOPID(imbr, intlist->i[i]);
	    if(opid == NULL)
		continue;

	    switch(opid->op)
	    {
	      case EDV_IMBR_OP_NONE:
		item[i] = item_ptr = NULL;
		break;

	      default:
                item[i] = item_ptr = ToolBarItemNew(
                    opid->toolbar_item_type,
                    opid->button_name,
		    opid->button_icon_data,
		    opid->tooltip,
                    opid->op,
                    EDVImbrOPCB, opid,
		    EDVImbrOPEnterCB, opid,                    
		    EDVImbrOPLeaveCB, opid
                );
                break;
	    }

	}

	/* Create new tool bar. */
	imbr->toolbar = ToolBarNew(
	    item, total_items, parent,
	    tb_display, tb_relief, FALSE
	);
	ToolBarMap(imbr->toolbar);


	/* Deallocate tool bar items list, it is no longer needed. */
        for(i = 0; i < total_items; i++)
	    ToolBarItemDelete(item[i]);
	g_free(item);
	item = NULL;
	total_items = 0;


	/* Need to tell toplevel to resize since tool bar may change size. */
	w = imbr->toplevel;
	if(w != NULL)
	    gtk_widget_queue_resize(w);
}

/*
 *	Recreates the directory ctree right click menu on the given
 *	image browser.
 */
void EDVImbrTListMenuRegenerate(edv_imbr_struct *imbr)
{
	if(imbr == NULL)
	    return;


	/* Destroy existing menu if it already exists. */
	if(1)
	{
	    GtkWidget **w;

#define DO_DESTROY_WIDGET       \
{ \
 if((*w) != NULL) \
 { \
  GtkWidget *tmp_w = *w; \
  (*w) = NULL; \
  gtk_widget_destroy(tmp_w); \
 } \
}

	    w = &imbr->tlist_menu;
	    DO_DESTROY_WIDGET

#undef DO_DESTROY_WIDGET
	}


	/* Create new menu. */
	if(1)
	{
	    gint op;
	    guint8 **icon_data = NULL;
	    GtkAccelGroup *accelgrp = NULL;	/* No acceleraters wanted. */
	    GtkWidget *menu, *w, **w_record, **fw_record;

#define DO_ADD_MENU_ITEM_LABEL  \
{ \
 edv_imbr_opid_struct *opid = EDVImbrMatchOPID(imbr, op); \
 if(opid != NULL) \
 { \
  w = GUIMenuItemCreate( \
   menu, GUI_MENU_ITEM_TYPE_LABEL, accelgrp, \
   (u_int8_t **)((icon_data != NULL) ? icon_data : opid->menu_icon_data), \
   opid->menu_name, \
   opid->accel_key, opid->accel_mods, \
   (void **)fw_record, \
   (gpointer)opid, EDVImbrMenuItemCB \
  ); \
  GUISetMenuItemCrossingCB( \
   w, \
   (gpointer)EDVImbrMenuItemEnterCB, (gpointer)opid, \
   (gpointer)EDVImbrMenuItemLeaveCB, (gpointer)opid \
  ); \
  if(w_record != NULL) *w_record = w; \
 } \
}

#define DO_ADD_MENU_SEP \
{ \
 w = GUIMenuItemCreate( \
  menu, GUI_MENU_ITEM_TYPE_SEPARATOR, NULL, \
  NULL, NULL, 0, 0, NULL, \
  NULL, NULL \
 ); \
}

	    /* Create menu. */
	    imbr->tlist_menu = menu = GUIMenuCreate();

            op = EDV_IMBR_OP_OPEN;
            w_record = &imbr->tlist_open_mi;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_LABEL

            op = EDV_IMBR_OP_OPEN_WITH;
            w_record = &imbr->tlist_openwith_mi;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_LABEL

            DO_ADD_MENU_SEP

            op = EDV_IMBR_OP_MOUNT;
            w_record = &imbr->tlist_mount_mi;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_LABEL

	    DO_ADD_MENU_SEP
	    imbr->tlist_mount_misep = w;

            op = EDV_IMBR_OP_MOVE;
            w_record = &imbr->tlist_move_mi;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_LABEL

            op = EDV_IMBR_OP_COPY;
            w_record = &imbr->tlist_copy_mi;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_LABEL

            op = EDV_IMBR_OP_LINK;
            w_record = &imbr->tlist_link_mi;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_LABEL

            op = EDV_IMBR_OP_RENAME;
            w_record = &imbr->tlist_rename_mi;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_LABEL

            op = EDV_IMBR_OP_CHMOD;
            w_record = &imbr->tlist_chmod_mi;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_LABEL

            op = EDV_IMBR_OP_CHOWN;
            w_record = &imbr->tlist_chown_mi;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_LABEL

            op = EDV_IMBR_OP_DELETE;
            w_record = &imbr->tlist_delete_mi;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_LABEL

            DO_ADD_MENU_SEP

            op = EDV_IMBR_OP_PROPERTIES;
            w_record = &imbr->tlist_properties_mi;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_LABEL

#undef DO_ADD_MENU_ITEM_LABEL
#undef DO_ADD_MENU_SEP
	}
}


/*
 *	Sets values on the given image browser to the configuration list
 *	on the core structure of the given image browser.
 */
void EDVImbrSyncConfiguration(edv_imbr_struct *imbr)
{
	GtkWidget *w;
	edv_core_struct *core_ptr;


	if(imbr == NULL)
	    return;

	core_ptr = (edv_core_struct *)imbr->core_ptr;
	if(core_ptr == NULL)
	    return;

	/* Get size of toplevel. */
	w = imbr->toplevel;
	if(w != NULL)
	{
	    GdkWindow *window = w->window;
	    gint x = 0, y = 0, width, height;


	    if(window != NULL)
		gdk_window_get_root_origin(window, &x, &y);
            EDVCFGItemListSetValueI(
                core_ptr->cfg_list, EDV_CFG_PARM_IMBR_X,
                x, FALSE
            );
            EDVCFGItemListSetValueI(
                core_ptr->cfg_list, EDV_CFG_PARM_IMBR_Y,
                y, FALSE
            );
	    gtk_widget_set_uposition(w, x, y);

	    width = w->allocation.width;
	    height = w->allocation.height;
	    EDVCFGItemListSetValueI(
		core_ptr->cfg_list, EDV_CFG_PARM_IMBR_WIDTH,
		width, FALSE
	    );
            EDVCFGItemListSetValueI(
                core_ptr->cfg_list, EDV_CFG_PARM_IMBR_HEIGHT,
                height, FALSE
            );
	    gtk_widget_set_usize(w, width, height);
	}

	/* Get position of main paned. */
        w = imbr->main_paned;
        if(w != NULL)
        {
	    gint width = GTK_PANED(w)->child1_size;
            EDVCFGItemListSetValueI(
                core_ptr->cfg_list, EDV_CFG_PARM_IMBR_PANED_POSITION,
                width, FALSE
            );
        }

	/* Set new startup directory as the last directory of this
	 * window's location.
	 */
	if(EDVCFGItemListGetValueI(
	    core_ptr->cfg_list, EDV_CFG_PARM_RECORD_START_UP_DIR
	))
	    EDVCFGItemListSetValueS(
		core_ptr->cfg_list, EDV_CFG_PARM_DIR_START_UP,
		EDVImbrCurrentLocation(imbr), FALSE
	    );


}

/*
 *	Resets the loading process timeout callback and initializes
 *	loading values.
 */
void EDVImbrQueueLoadingProcess(edv_imbr_struct *imbr)
{
	if(imbr == NULL)
	    return;

	/* Remove previous loading timeout callback as needed. */
	if(imbr->loading_tocb != (guint)-1)
	{
	    gtk_timeout_remove(imbr->loading_tocb);
	    imbr->loading_tocb = (guint)-1;
	}

	/* Clear stop count. */
	imbr->stop_count = 0;

	/* Set new loading timeout. */
	imbr->loading_tocb = gtk_timeout_add(
	    100,
	    (GtkFunction)EDVImbrLoadingTOCB,
	    imbr
	);
}


/*
 *	Clears the thumbs list and resets all loading procedure values.
 *	Then loads the listing for the new directory listing if path
 *	is valid and loading procedure values will be initialized again.
 */
void EDVImbrSelectPath(edv_imbr_struct *imbr, const gchar *path)
{
	gchar *dpath;


	if((imbr == NULL) || (path == NULL))
	    return;

	dpath = g_strdup(path);
	if(dpath == NULL)
	    return;


        /* Remove previous loading timeout callback as needed. */
        if(imbr->loading_tocb != (guint)-1)
        {
            gtk_timeout_remove(imbr->loading_tocb);
            imbr->loading_tocb = (guint)-1;
        }


	/* Update title and location bar. */
	EDVImbrSetTitle(imbr, dpath);
	EDVImbrSetLocation(imbr, dpath, FALSE);
	EDVImbrUpdateLocationIcon(imbr, dpath);


	/* Clear thumbs list and get new listing. */
	EDVImbrTListDoUpdate(imbr, dpath, TRUE);

	/* Set loading progress timeout callback and reset loading
	 * progress values.
	 */
	EDVImbrQueueLoadingProcess(imbr);

	g_free(dpath);
	dpath = NULL;
}



/*
 *	Creates a new standard list of image browser op ids.
 */
edv_imbr_opid_struct **EDVImbrOPIDListNew(
        edv_imbr_struct *imbr,		/* Can be NULL. */
        gint *total
)
{
	gint i;
	edv_imbr_opid_struct **list, *opid;
        gint op, toolbar_item_type;
	gbool allow_multiple;
        void (*func_cb)(gpointer);
        void (*enter_func_cb)(gpointer);
        const gchar *button_name, *menu_name, *tooltip;
        gint accel_key;
        guint accel_mods;
	guint8 **button_icon_data, **menu_icon_data;


	if(total == NULL)
	    return(NULL);

	*total = 0;
	list = NULL;

#define DO_APPEND_OPID	\
{ \
 i = *total; \
 *total = i + 1; \
 list = (edv_imbr_opid_struct **)g_realloc( \
  list, \
  (*total) * sizeof(edv_imbr_opid_struct *) \
 ); \
 if(list == NULL) \
 { \
  *total = 0; \
  return(list); \
 } \
 else \
 { \
  list[i] = opid = (edv_imbr_opid_struct *)g_malloc0( \
   sizeof(edv_imbr_opid_struct) \
  ); \
  if(opid != NULL) \
  { \
   opid->imbr = imbr; \
   opid->op = op; \
   opid->toolbar_item_type = toolbar_item_type; \
   opid->allow_multiple = allow_multiple; \
   opid->func_cb = func_cb; \
   opid->enter_func_cb = enter_func_cb; \
   opid->button_name = (button_name != NULL) ? g_strdup(button_name) : NULL; \
   opid->menu_name = (menu_name != NULL) ? g_strdup(menu_name) : NULL; \
   opid->tooltip = (tooltip != NULL) ? g_strdup(tooltip) : NULL; \
   opid->accel_key = accel_key; \
   opid->accel_mods = accel_mods; \
   opid->button_icon_data = button_icon_data; \
   opid->menu_icon_data = menu_icon_data; \
  } \
 } \
}

	func_cb = EDVImbrOPCB;
	enter_func_cb = EDVImbrOPEnterCB;


        op = EDV_IMBR_OP_SEPARATOR;
	toolbar_item_type = TOOLBAR_ITEM_SEPARATOR;
	allow_multiple = TRUE;
	button_icon_data = NULL;
        menu_icon_data = NULL;
        button_name = "Separator";
        menu_name = "Separator";
        tooltip = "Separator";
	accel_key = 0;
	accel_mods = 0;
        DO_APPEND_OPID


        op = EDV_IMBR_OP_CLOSE;
        toolbar_item_type = TOOLBAR_ITEM_BUTTON;
        allow_multiple = FALSE;
        button_icon_data = (guint8 **)icon_close_20x20_xpm;
	menu_icon_data = button_icon_data;
        button_name = "Close";
        menu_name = "Close";
        tooltip = "Close this window";
        accel_key = 'w';
        accel_mods = GDK_CONTROL_MASK;
        DO_APPEND_OPID

        op = EDV_IMBR_OP_EXIT;
        toolbar_item_type = TOOLBAR_ITEM_BUTTON;
        allow_multiple = FALSE;
	button_icon_data = (guint8 **)icon_exit_20x20_xpm;
        menu_icon_data = button_icon_data;
        button_name = "Exit";
        menu_name = "Exit";
        tooltip = "Exit this application";
        accel_key = 0;
        accel_mods = 0;
        DO_APPEND_OPID


        op = EDV_IMBR_OP_WRITE_PROTECT;
        toolbar_item_type = TOOLBAR_ITEM_BUTTON;
        allow_multiple = FALSE;
	button_icon_data = (guint8 **)icon_secure_20x20_xpm;
        menu_icon_data = button_icon_data;
        button_name = "Protect";
        menu_name = "Write Protect";
        tooltip = "Enable/disable write protect";
        accel_key = 0;
        accel_mods = 0;
        DO_APPEND_OPID

        op = EDV_IMBR_OP_SYNC_DISKS;
        toolbar_item_type = TOOLBAR_ITEM_BUTTON;
        allow_multiple = FALSE;
        button_icon_data = (guint8 **)icon_save_20x20_xpm;
        menu_icon_data = button_icon_data;
        button_name = "Sync Disks";
        menu_name = "Sync Disks";
        tooltip = "Flush any pending data transfers to disk";
        accel_key = 0;
        accel_mods = 0;
        DO_APPEND_OPID


        op = EDV_IMBR_OP_NEW;
        toolbar_item_type = TOOLBAR_ITEM_BUTTON;
        allow_multiple = FALSE;
        button_icon_data = (guint8 **)icon_file_20x20_xpm;
        menu_icon_data = button_icon_data;
        button_name = "New";
        menu_name = "New";
        tooltip = "Create a new object";
        accel_key = 0;
        accel_mods = 0;
        DO_APPEND_OPID

	op = EDV_IMBR_OP_NEW_FILE;
        toolbar_item_type = TOOLBAR_ITEM_BUTTON;
        allow_multiple = FALSE;
        button_icon_data = (guint8 **)icon_file_20x20_xpm;
        menu_icon_data = button_icon_data;
	button_name = "New File";
	menu_name = "File";
	tooltip = "Create a new file object";
        accel_key = 0;
        accel_mods = 0;
	DO_APPEND_OPID

        op = EDV_IMBR_OP_NEW_DIRECTORY;
        toolbar_item_type = TOOLBAR_ITEM_BUTTON;
        allow_multiple = FALSE;
        button_icon_data = (guint8 **)icon_folder_closed_20x20_xpm;
        menu_icon_data = button_icon_data;
        button_name = "New Dir";
        menu_name = "Directory";
        tooltip = "Create a new directory object";
        accel_key = 0;
        accel_mods = 0;
        DO_APPEND_OPID

        op = EDV_IMBR_OP_NEW_LINK;
        toolbar_item_type = TOOLBAR_ITEM_BUTTON;
        allow_multiple = FALSE;
        button_icon_data = (guint8 **)icon_link2_20x20_xpm;
        menu_icon_data = button_icon_data;
        button_name = "New Lnk";
        menu_name = "Link";
        tooltip = "Create a new symbolic link object";
        accel_key = 0;
        accel_mods = 0;
        DO_APPEND_OPID

        op = EDV_IMBR_OP_NEW_FIFO;
        toolbar_item_type = TOOLBAR_ITEM_BUTTON;
        allow_multiple = FALSE;
        button_icon_data = (guint8 **)icon_pipe_20x20_xpm;
        menu_icon_data = button_icon_data;
        button_name = "New FIFO";
        menu_name = "FIFO Pipe";
        tooltip = "Create a new FIFO pipe object";
        accel_key = 0;
        accel_mods = 0;
        DO_APPEND_OPID

        op = EDV_IMBR_OP_NEW_DEVICE_BLOCK;
        toolbar_item_type = TOOLBAR_ITEM_BUTTON;
        allow_multiple = FALSE;
        button_icon_data = (guint8 **)icon_device_block_20x20_xpm;
        menu_icon_data = button_icon_data;
        button_name = "New BDev";
        menu_name = "Block Device";
        tooltip = "Create a new block (buffered) device node object";
        accel_key = 0;
        accel_mods = 0;
        DO_APPEND_OPID

        op = EDV_IMBR_OP_NEW_DEVICE_CHARACTER;
        toolbar_item_type = TOOLBAR_ITEM_BUTTON;
        allow_multiple = FALSE;
        button_icon_data = (guint8 **)icon_device_character_20x20_xpm;
        menu_icon_data = button_icon_data;
        button_name = "New CDev";
        menu_name = "Character Device";
        tooltip = "Create a new character (unbuffered) device node object";
        accel_key = 0;
        accel_mods = 0;
        DO_APPEND_OPID

        op = EDV_IMBR_OP_NEW_SOCKET;
        toolbar_item_type = TOOLBAR_ITEM_BUTTON;
        allow_multiple = FALSE;
        button_icon_data = (guint8 **)icon_socket_20x20_xpm;
        menu_icon_data = button_icon_data;
        button_name = "New Sock";
        menu_name = "Socket";
        tooltip = "Create a new socket object";
        accel_key = 0;
        accel_mods = 0;
        DO_APPEND_OPID


	op = EDV_IMBR_OP_OPEN;
        toolbar_item_type = TOOLBAR_ITEM_BUTTON;
        allow_multiple = FALSE;
        button_icon_data = (guint8 **)icon_open_20x20_xpm;
        menu_icon_data = button_icon_data;
        button_name = "Open";
        menu_name = "Open";
        tooltip = "Open selected object";
        accel_key = 0;
        accel_mods = 0;
        DO_APPEND_OPID

        op = EDV_IMBR_OP_OPEN_WITH;
        toolbar_item_type = TOOLBAR_ITEM_BUTTON;
        allow_multiple = FALSE;
	button_icon_data = (guint8 **)icon_open_20x20_xpm;
        menu_icon_data = NULL;
        button_name = "Open With";
        menu_name = "Open With...";
        tooltip = "Open selected object using a specific method";
        accel_key = 0;
        accel_mods = 0;
        DO_APPEND_OPID


	op = EDV_IMBR_OP_GOTO_PARENT;
        toolbar_item_type = TOOLBAR_ITEM_BUTTON;
        allow_multiple = FALSE;
        button_icon_data = (guint8 **)icon_folder_parent_20x20_xpm;
        menu_icon_data = button_icon_data;
        button_name = "Parent";
        menu_name = "Go To Parent";
        tooltip = "Go to the parent directory";
        accel_key = 0;
        accel_mods = 0;
        DO_APPEND_OPID

        op = EDV_IMBR_OP_MOUNT;
        toolbar_item_type = TOOLBAR_ITEM_BUTTON;
        allow_multiple = FALSE;
        button_icon_data = NULL;
        menu_icon_data = button_icon_data;
        button_name = "Mount";
        menu_name = "Mount";
        tooltip = "Mount/unmount selected mount point directory";
        accel_key = 0;
        accel_mods = 0;
        DO_APPEND_OPID

        op = EDV_IMBR_OP_MOVE;
        toolbar_item_type = TOOLBAR_ITEM_BUTTON;
        allow_multiple = FALSE;
        button_icon_data = (guint8 **)icon_move_file_20x20_xpm;
        menu_icon_data = button_icon_data;
        button_name = "Move";
        menu_name = "Move...";
        tooltip = "Move selected object(s)";
        accel_key = GDK_F7;
        accel_mods = 0;
        DO_APPEND_OPID

        op = EDV_IMBR_OP_COPY;
        toolbar_item_type = TOOLBAR_ITEM_BUTTON;
        allow_multiple = FALSE;
        button_icon_data = (guint8 **)icon_copy_file_20x20_xpm;
        menu_icon_data = button_icon_data;
        button_name = "Copy";
        menu_name = "Copy...";
        tooltip = "Copy selected object(s)";
        accel_key = GDK_F8;
        accel_mods = 0;
        DO_APPEND_OPID

        op = EDV_IMBR_OP_LINK;
        toolbar_item_type = TOOLBAR_ITEM_BUTTON;
        allow_multiple = FALSE;
        button_icon_data = (guint8 **)icon_link2_20x20_xpm;
        menu_icon_data = button_icon_data;
        button_name = "Link";
        menu_name = "Link...";
        tooltip = "Link selected object(s)";
        accel_key = GDK_F6;
        accel_mods = 0;
        DO_APPEND_OPID

        op = EDV_IMBR_OP_RENAME;
        toolbar_item_type = TOOLBAR_ITEM_BUTTON;
        allow_multiple = FALSE;
        button_icon_data = (guint8 **)icon_rename_20x20_xpm;
        menu_icon_data = button_icon_data;
        button_name = "Rename";
        menu_name = "Rename";
        tooltip = "Rename selected object";
        accel_key = 'r';
        accel_mods = GDK_CONTROL_MASK;
        DO_APPEND_OPID

        op = EDV_IMBR_OP_CHMOD;
        toolbar_item_type = TOOLBAR_ITEM_BUTTON;
        allow_multiple = FALSE;
        button_icon_data = (guint8 **)icon_chmod_20x20_xpm;
        menu_icon_data = button_icon_data;
        button_name = "ChMod";
        menu_name = "Change Mode...";
        tooltip = "Change selected objects' permissions";
        accel_key = GDK_F9;
        accel_mods = 0;
        DO_APPEND_OPID

        op = EDV_IMBR_OP_CHOWN;
        toolbar_item_type = TOOLBAR_ITEM_BUTTON;
        allow_multiple = FALSE;
        button_icon_data = (guint8 **)icon_owned_20x20_xpm;
        menu_icon_data = button_icon_data;
        button_name = "Chown";
        menu_name = "Chown...";
        tooltip = "Chown selected object(s)";
        accel_key = GDK_F10;
        accel_mods = 0;
        DO_APPEND_OPID

        op = EDV_IMBR_OP_DELETE;
        toolbar_item_type = TOOLBAR_ITEM_BUTTON;
        allow_multiple = FALSE;
        button_icon_data = (guint8 **)icon_cancel_20x20_xpm;
        menu_icon_data = button_icon_data;
        button_name = "Delete";
        menu_name = "Delete";
        tooltip = "Delete selected object(s)";
        accel_key = GDK_Delete;
        accel_mods = GDK_CONTROL_MASK;
        DO_APPEND_OPID

        op = EDV_IMBR_OP_SELECT_ALL;
        toolbar_item_type = TOOLBAR_ITEM_BUTTON;
        allow_multiple = FALSE;
        button_icon_data = NULL;
        menu_icon_data = button_icon_data;
        button_name = "Select All";
        menu_name = "Select All";
        tooltip = "Select all objects";
        accel_key = 'a';
        accel_mods = GDK_CONTROL_MASK;
        DO_APPEND_OPID

        op = EDV_IMBR_OP_UNSELECT_ALL;
        toolbar_item_type = TOOLBAR_ITEM_BUTTON;
        allow_multiple = FALSE;
        button_icon_data = NULL;
        menu_icon_data = button_icon_data;
        button_name = "Unselect All";
        menu_name = "Unselect All";
        tooltip = "Unselect all objects";
        accel_key = 'u';
        accel_mods = GDK_CONTROL_MASK;
        DO_APPEND_OPID

        op = EDV_IMBR_OP_PROPERTIES;
        toolbar_item_type = TOOLBAR_ITEM_BUTTON;
        allow_multiple = FALSE;
        button_icon_data = (guint8 **)icon_properties2_20x20_xpm;
        menu_icon_data = button_icon_data;
        button_name = "Properties";
        menu_name = "Properties...";
        tooltip = "Modify selected object's properties";
        accel_key = GDK_Return;
        accel_mods = GDK_CONTROL_MASK;
        DO_APPEND_OPID

        op = EDV_IMBR_OP_FIND;
        toolbar_item_type = TOOLBAR_ITEM_BUTTON;
        allow_multiple = FALSE;
        button_icon_data = (guint8 **)icon_search_20x20_xpm;
        menu_icon_data = button_icon_data;
        button_name = "Find";
        menu_name = "Find...";
        tooltip = "Find object(s) by specified attributes";
        accel_key = 'f';
        accel_mods = GDK_CONTROL_MASK;
        DO_APPEND_OPID


        op = EDV_IMBR_OP_REFRESH;
        toolbar_item_type = TOOLBAR_ITEM_BUTTON;
        allow_multiple = FALSE;
        button_icon_data = (guint8 **)icon_reload_20x20_xpm;
        menu_icon_data = button_icon_data;
        button_name = "Refresh";
        menu_name = "Refresh";
        tooltip = "Refresh selected directory's listing";
        accel_key = GDK_F5;
        accel_mods = 0;
        DO_APPEND_OPID

        op = EDV_IMBR_OP_REFRESH_ALL;
        toolbar_item_type = TOOLBAR_ITEM_BUTTON;
        allow_multiple = FALSE;
        button_icon_data = (guint8 **)icon_reload_20x20_xpm;
        menu_icon_data = button_icon_data;
        button_name = "Refresh All";
        menu_name = "Refresh All";
        tooltip = "Refresh all listings";
        accel_key = 0;
        accel_mods = 0;
        DO_APPEND_OPID

        op = EDV_IMBR_OP_STOP;
        toolbar_item_type = TOOLBAR_ITEM_BUTTON;
        allow_multiple = FALSE;
        button_icon_data = (guint8 **)icon_stop_20x20_xpm;
        menu_icon_data = button_icon_data;
        button_name = "Stop";
        menu_name = "Stop Loading";
        tooltip = "Stop the loading process";
        accel_key = 0;
        accel_mods = 0;
        DO_APPEND_OPID

        op = EDV_IMBR_OP_CONTINUE;
        toolbar_item_type = TOOLBAR_ITEM_BUTTON;
        allow_multiple = FALSE;
        button_icon_data = (guint8 **)icon_run_20x20_xpm;
        menu_icon_data = button_icon_data;
        button_name = "Continue";
        menu_name = "Continue Loading";
        tooltip = "Continue the loading process";
        accel_key = 0;
        accel_mods = 0;
        DO_APPEND_OPID

        op = EDV_IMBR_OP_SHOW_TOOL_BAR;
        toolbar_item_type = TOOLBAR_ITEM_TOGGLE_BUTTON;
        allow_multiple = FALSE;
        button_icon_data = (guint8 **)icon_tool_bar_20x20_xpm;
        menu_icon_data = button_icon_data;
        button_name = "ToolBar";
        menu_name = "Tool Bar";
        tooltip = "Show/hide tool bar";
        accel_key = 0;
        accel_mods = 0;
        DO_APPEND_OPID

        op = EDV_IMBR_OP_SHOW_LOCATION_BAR;
        toolbar_item_type = TOOLBAR_ITEM_TOGGLE_BUTTON;
        allow_multiple = FALSE;
        button_icon_data = (guint8 **)icon_tool_bar_20x20_xpm;
        menu_icon_data = button_icon_data;
        button_name = "LocBar";
        menu_name = "Location Bar";
        tooltip = "Show/hide location bar";
        accel_key = 0;
        accel_mods = 0;
        DO_APPEND_OPID

        op = EDV_IMBR_OP_SHOW_MOUNT_BAR;
        toolbar_item_type = TOOLBAR_ITEM_TOGGLE_BUTTON;
        allow_multiple = FALSE;
        button_icon_data = (guint8 **)icon_tool_bar_20x20_xpm;
        menu_icon_data = button_icon_data;
        button_name = "MountBar";
        menu_name = "Mount Bar";
        tooltip = "Show/hide mount bar";
        accel_key = 0;
        accel_mods = 0;
        DO_APPEND_OPID

        op = EDV_IMBR_OP_SHOW_FIND_BAR;
        toolbar_item_type = TOOLBAR_ITEM_TOGGLE_BUTTON;
        allow_multiple = FALSE;
        button_icon_data = (guint8 **)icon_tool_bar_20x20_xpm;
        menu_icon_data = button_icon_data;
        button_name = "FindBar";
        menu_name = "Find Bar";
        tooltip = "Show/hide find bar";
        accel_key = 0;
        accel_mods = 0;
        DO_APPEND_OPID

        op = EDV_IMBR_OP_SHOW_STATUS_BAR;
        toolbar_item_type = TOOLBAR_ITEM_TOGGLE_BUTTON;
        allow_multiple = FALSE;
        button_icon_data = (guint8 **)icon_status_bar_20x20_xpm;
        menu_icon_data = button_icon_data;
        button_name = "StatusBar";
        menu_name = "Status Bar";
        tooltip = "Show/hide status bar";
        accel_key = 0;
        accel_mods = 0;
        DO_APPEND_OPID

        op = EDV_IMBR_OP_MIME_TYPES;
        toolbar_item_type = TOOLBAR_ITEM_BUTTON;
        allow_multiple = FALSE;
        button_icon_data = (guint8 **)icon_properties2_20x20_xpm;
        menu_icon_data = button_icon_data;
        button_name = "MIMETypes";
        menu_name = "MIME Types...";
        tooltip = "MIME Types list";
        accel_key = 'm';
        accel_mods = GDK_CONTROL_MASK;
        DO_APPEND_OPID

        op = EDV_IMBR_OP_DEVICES;
        toolbar_item_type = TOOLBAR_ITEM_BUTTON;
        allow_multiple = FALSE;
        button_icon_data = (guint8 **)icon_device_misc_20x20_xpm;
        menu_icon_data = button_icon_data;
        button_name = "Devices";
        menu_name = "Devices...";
        tooltip = "File system devices list";
        accel_key = 'd';
        accel_mods = GDK_CONTROL_MASK;
        DO_APPEND_OPID


        op = EDV_IMBR_OP_NEW_BROWSER;
        toolbar_item_type = TOOLBAR_ITEM_BUTTON;
        allow_multiple = FALSE;
        button_icon_data = (guint8 **)endeavour_20x20_xpm;
        menu_icon_data = button_icon_data;
        button_name = "FBrowser";
        menu_name = "New File Browser";
        tooltip = "Create a new file browser";
        accel_key = 0;
        accel_mods = 0;
        DO_APPEND_OPID

        op = EDV_IMBR_OP_NEW_IMBR;
        toolbar_item_type = TOOLBAR_ITEM_BUTTON;
        allow_multiple = FALSE;
        button_icon_data = (guint8 **)image_browser_20x20_xpm;
        menu_icon_data = button_icon_data;
        button_name = "IBrowser";
        menu_name = "New Image Browser";
        tooltip = "Create a new image browser";
        accel_key = 0;
        accel_mods = 0;
        DO_APPEND_OPID

        op = EDV_IMBR_OP_NEW_ARCHIVER;
        toolbar_item_type = TOOLBAR_ITEM_BUTTON;
        allow_multiple = FALSE;
        button_icon_data = (guint8 **)archiver_20x20_xpm;
        menu_icon_data = button_icon_data;
        button_name = "Archiver";
        menu_name = "New Archiver";
        tooltip = "Create a new archiver";
        accel_key = 0;
        accel_mods = 0;
        DO_APPEND_OPID

        op = EDV_IMBR_OP_RECYCLE_BIN;
        toolbar_item_type = TOOLBAR_ITEM_BUTTON;
        allow_multiple = FALSE;
        button_icon_data = (guint8 **)icon_trash_20x20_xpm;
        menu_icon_data = button_icon_data;
        button_name = "RecBin";
        menu_name = "Recycle Bin";
        tooltip = "Recycle bin (trash can)";
        accel_key = 0;
        accel_mods = 0;
        DO_APPEND_OPID


        op = EDV_IMBR_OP_OPTIONS;
        toolbar_item_type = TOOLBAR_ITEM_BUTTON;
        allow_multiple = FALSE;
        button_icon_data = (guint8 **)icon_options2_20x20_xpm;
        menu_icon_data = button_icon_data;
        button_name = "Options";
        menu_name = "Options...";
        tooltip = "Modify this application's options";
        accel_key = 'o';
        accel_mods = GDK_CONTROL_MASK;
        DO_APPEND_OPID

        op = EDV_IMBR_OP_CUSTOMIZE;
        toolbar_item_type = TOOLBAR_ITEM_BUTTON;
        allow_multiple = FALSE;
        button_icon_data = (guint8 **)icon_customize_20x20_xpm;
        menu_icon_data = button_icon_data;
        button_name = "Custom";
        menu_name = "Customize...";
        tooltip = "Customize the look and feel of this application";
        accel_key = 't';
        accel_mods = GDK_CONTROL_MASK;
        DO_APPEND_OPID


        op = EDV_IMBR_OP_HELP_CONTENTS;
        toolbar_item_type = TOOLBAR_ITEM_BUTTON;
        allow_multiple = FALSE;
        button_icon_data = (guint8 **)icon_help_20x20_xpm;
        menu_icon_data = NULL;
        button_name = "Help";
        menu_name = "Contents";
        tooltip = "List of all help contents";
        accel_key = GDK_F1;
        accel_mods = 0;
        DO_APPEND_OPID

        op = EDV_IMBR_OP_HELP_KEYS_LIST;
        toolbar_item_type = TOOLBAR_ITEM_BUTTON;
        allow_multiple = FALSE;
        button_icon_data = (guint8 **)icon_help_20x20_xpm;
        menu_icon_data = NULL;
        button_name = "Help Keys";
        menu_name = "Keys List";
        tooltip = "List of all keyboard key operations";
        accel_key = 0;
        accel_mods = 0;
        DO_APPEND_OPID

        op = EDV_IMBR_OP_HELP_MIME_TYPES;
        toolbar_item_type = TOOLBAR_ITEM_BUTTON;
        allow_multiple = FALSE;
        button_icon_data = (guint8 **)icon_help_20x20_xpm;
        menu_icon_data = NULL;
        button_name = "Help MIME";
        menu_name = "Mime Types";
        tooltip = "Help on setting up MIME Types";
        accel_key = 0;
        accel_mods = 0;
        DO_APPEND_OPID

        op = EDV_IMBR_OP_HELP_DEVICES;
        toolbar_item_type = TOOLBAR_ITEM_BUTTON;
        allow_multiple = FALSE;
        button_icon_data = (guint8 **)icon_help_20x20_xpm;
        menu_icon_data = NULL;
        button_name = "Help Dev";
        menu_name = "Devices";
        tooltip = "Help on setting up devices";
        accel_key = 0;
        accel_mods = 0;
        DO_APPEND_OPID

        op = EDV_IMBR_OP_HELP_COMMON_OPERATIONS;
        toolbar_item_type = TOOLBAR_ITEM_BUTTON;
        allow_multiple = FALSE;
        button_icon_data = (guint8 **)icon_help_20x20_xpm;
        menu_icon_data = NULL;
        button_name = "Help Op";
        menu_name = "Common Operations";
        tooltip = "Help on the most common operations";
        accel_key = 0;
        accel_mods = 0;
        DO_APPEND_OPID

        op = EDV_IMBR_OP_HELP_ABOUT;
        toolbar_item_type = TOOLBAR_ITEM_BUTTON;
        allow_multiple = FALSE;
        button_icon_data = (guint8 **)icon_about_20x20_xpm;
        menu_icon_data = (guint8 **)icon_about_20x20_xpm;
        button_name = "About";
        menu_name = "About";
        tooltip = "About this application";
        accel_key = 0;
        accel_mods = 0;
        DO_APPEND_OPID


#undef DO_APPEND_OPID

	return(list);
}

/*
 *	Deallocates the given list of image browser op ids.
 */
void EDVImbrOPIDListDelete(
        edv_imbr_opid_struct **list, gint total
)
{
	gint i;
	edv_imbr_opid_struct *opid;


	if(list == NULL)
	    return;

        for(i = 0; i < total; i++)
        {
            opid = list[i];
            if(opid == NULL)
                continue;

            g_free(opid->button_name);
            g_free(opid->menu_name);
            g_free(opid->tooltip);

            g_free(opid);
        }
        g_free(list);
	list = NULL;
	total = 0;
}


/*
 *	Creates the menu bar for the given image browser.
 *
 *	This function should be called by EDVImbrNew().
 */
static void EDVImbrCreateMenuBar(
	edv_imbr_struct *imbr, GtkWidget *parent
)
{
	gint op;
	guint8 **icon_data = NULL;
        GtkAccelGroup *accelgrp;
        GtkWidget *menu_bar, *menu, *submenu, *w, **w_record, **fw_record;


	/* Get keyboard accelerator group. */
	accelgrp = imbr->accelgrp;

        /* Create menu bar. */
        menu_bar = (GtkWidget *)GUIMenuBarCreate(NULL);
	imbr->menu_bar = menu_bar;
        gtk_container_add(GTK_CONTAINER(parent), menu_bar);
        gtk_widget_show(menu_bar);

#define DO_ADD_MENU_ITEM_LABEL  \
{ \
 edv_imbr_opid_struct *opid = EDVImbrMatchOPID(imbr, op); \
 if(opid != NULL) \
 { \
  w = GUIMenuItemCreate( \
   menu, GUI_MENU_ITEM_TYPE_LABEL, accelgrp, \
   (u_int8_t **)((icon_data != NULL) ? icon_data : opid->menu_icon_data), \
   opid->menu_name, \
   opid->accel_key, opid->accel_mods, \
   (void **)fw_record, \
   (gpointer)opid, EDVImbrMenuItemCB \
  ); \
  GUISetMenuItemCrossingCB( \
   w, \
   (gpointer)EDVImbrMenuItemEnterCB, (gpointer)opid, \
   (gpointer)EDVImbrMenuItemLeaveCB, (gpointer)opid \
  ); \
  if(w_record != NULL) *w_record = w; \
 } \
}
#define DO_ADD_MENU_ITEM_SUBMENU        \
{ \
 edv_imbr_opid_struct *opid = EDVImbrMatchOPID(imbr, op); \
 if(opid != NULL) \
 { \
  w = GUIMenuItemCreate( \
   menu, GUI_MENU_ITEM_TYPE_SUBMENU, accelgrp, \
   (u_int8_t **)((icon_data != NULL) ? icon_data : opid->menu_icon_data), \
   opid->menu_name, \
   opid->accel_key, opid->accel_mods, \
   (void **)fw_record, \
   (gpointer)opid, NULL \
  ); \
  GUISetMenuItemCrossingCB( \
   w, \
   (gpointer)EDVImbrMenuItemEnterCB, (gpointer)opid, \
   (gpointer)EDVImbrMenuItemLeaveCB, (gpointer)opid \
  ); \
  if(w != NULL) \
   GUIMenuItemSetSubMenu(w, submenu); \
  if(w_record != NULL) *w_record = w; \
 } \
}

#define DO_ADD_MENU_ITEM_CHECK  \
{ \
 edv_imbr_opid_struct *opid = EDVImbrMatchOPID(imbr, op); \
 if(opid != NULL) \
 { \
  w = GUIMenuItemCreate( \
   menu, GUI_MENU_ITEM_TYPE_CHECK, accelgrp, \
   (u_int8_t **)((icon_data != NULL) ? icon_data : opid->menu_icon_data), \
   opid->menu_name, \
   opid->accel_key, opid->accel_mods, \
   (void **)fw_record, \
   (gpointer)opid, EDVImbrMenuItemCB \
  ); \
  GUISetMenuItemCrossingCB( \
   w, \
   (gpointer)EDVImbrMenuItemEnterCB, (gpointer)opid, \
   (gpointer)EDVImbrMenuItemLeaveCB, (gpointer)opid \
  ); \
  if(w_record != NULL) *w_record = w; \
 } \
}

#define DO_ADD_MENU_SEP \
{ \
 w = GUIMenuItemCreate( \
  menu, GUI_MENU_ITEM_TYPE_SEPARATOR, NULL, \
  NULL, NULL, 0, 0, NULL, \
  NULL, NULL \
 ); \
}

	/* Create new object submenu. */
        menu = GUIMenuCreate();
	if(menu != NULL)
	{
            op = EDV_IMBR_OP_NEW_FILE;
            w_record = &imbr->new_file_mi;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_LABEL

	    op = EDV_IMBR_OP_NEW_DIRECTORY;
	    w_record = &imbr->new_directory_mi;
	    fw_record = NULL;
	    DO_ADD_MENU_ITEM_LABEL

            op = EDV_IMBR_OP_NEW_LINK;
            w_record = &imbr->new_link_mi;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_LABEL

            op = EDV_IMBR_OP_NEW_FIFO;
            w_record = &imbr->new_fifo_mi;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_LABEL

            op = EDV_IMBR_OP_NEW_DEVICE_BLOCK;
            w_record = &imbr->new_device_block_mi;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_LABEL

            op = EDV_IMBR_OP_NEW_DEVICE_CHARACTER;
            w_record = &imbr->new_device_character_mi;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_LABEL

            op = EDV_IMBR_OP_NEW_SOCKET;
            w_record = &imbr->new_socket_mi;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_LABEL
	}
	imbr->new_object_submenu = submenu = menu;

        /* Create file menu. */
        menu = GUIMenuCreateTearOff();
        if(menu != NULL)
        {
	    op = EDV_IMBR_OP_OPEN;
	    w_record = &imbr->open_mi;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_LABEL

            op = EDV_IMBR_OP_OPEN_WITH;
            w_record = &imbr->openwith_mi;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_LABEL

	    DO_ADD_MENU_SEP

	    op = EDV_IMBR_OP_NEW;
	    w_record = &imbr->new_object_submi;
	    fw_record = NULL;
	    DO_ADD_MENU_ITEM_SUBMENU

	    DO_ADD_MENU_SEP

            op = EDV_IMBR_OP_SYNC_DISKS;
            w_record = &imbr->sync_disks_mi;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_LABEL

            DO_ADD_MENU_SEP

            op = EDV_IMBR_OP_CLOSE;
            w_record = &imbr->close_mi;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_LABEL
        }
        GUIMenuAddToMenuBar(
            menu_bar, menu,
            "File",
            GUI_MENU_BAR_ALIGN_LEFT
        );


        /* Create edit menu. */
        menu = GUIMenuCreateTearOff();
        if(menu != NULL)
        {
            op = EDV_IMBR_OP_MOVE;
            w_record = &imbr->move_mi;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_LABEL

            op = EDV_IMBR_OP_COPY;
            w_record = &imbr->copy_mi;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_LABEL

            op = EDV_IMBR_OP_LINK;
            w_record = &imbr->link_mi;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_LABEL

            op = EDV_IMBR_OP_RENAME;
            w_record = &imbr->rename_mi;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_LABEL

            op = EDV_IMBR_OP_CHMOD;
            w_record = &imbr->chmod_mi;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_LABEL

            op = EDV_IMBR_OP_CHOWN;
            w_record = &imbr->chown_mi;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_LABEL

            op = EDV_IMBR_OP_DELETE;
            w_record = &imbr->delete_mi;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_LABEL

            DO_ADD_MENU_SEP

            op = EDV_IMBR_OP_SELECT_ALL;
            w_record = &imbr->select_all_mi;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_LABEL

            op = EDV_IMBR_OP_UNSELECT_ALL;
            w_record = &imbr->unselect_all_mi;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_LABEL

            op = EDV_IMBR_OP_FIND;
            w_record = &imbr->find_mi;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_LABEL

	    DO_ADD_MENU_SEP

            op = EDV_IMBR_OP_PROPERTIES;
            w_record = &imbr->properties_mi;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_LABEL
        }
        GUIMenuAddToMenuBar(
            menu_bar, menu,
            "Edit",
            GUI_MENU_BAR_ALIGN_LEFT
        );


        /* Create view menu. */
        menu = GUIMenuCreateTearOff();
        if(menu != NULL)
        {
            op = EDV_IMBR_OP_STOP;
            w_record = &imbr->stop_mi;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_LABEL

            op = EDV_IMBR_OP_CONTINUE;
            w_record = &imbr->continue_mi;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_LABEL

            DO_ADD_MENU_SEP

            op = EDV_IMBR_OP_REFRESH;
            w_record = &imbr->refresh_mi;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_LABEL

            op = EDV_IMBR_OP_REFRESH_ALL;
            w_record = &imbr->refresh_all_mi;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_LABEL

	    DO_ADD_MENU_SEP

            op = EDV_IMBR_OP_SHOW_TOOL_BAR;
            w_record = &imbr->show_tool_bar_micheck;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_CHECK

            op = EDV_IMBR_OP_SHOW_LOCATION_BAR;
            w_record = &imbr->show_location_bar_micheck;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_CHECK

            op = EDV_IMBR_OP_SHOW_MOUNT_BAR;
            w_record = &imbr->show_mount_bar_micheck;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_CHECK

            op = EDV_IMBR_OP_SHOW_FIND_BAR;
            w_record = &imbr->show_find_bar_micheck;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_CHECK

            op = EDV_IMBR_OP_SHOW_STATUS_BAR;
            w_record = &imbr->show_status_bar_micheck;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_CHECK

            DO_ADD_MENU_SEP

            op = EDV_IMBR_OP_MIME_TYPES;
            w_record = &imbr->mime_types_mi;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_LABEL
        }
        GUIMenuAddToMenuBar(
            menu_bar, menu,
            "View",
            GUI_MENU_BAR_ALIGN_LEFT
        );


        /* Create device menu. */
        menu = GUIMenuCreateTearOff();
        if(menu != NULL)
        {
            op = EDV_IMBR_OP_MOUNT;
            w_record = &imbr->mount_mi;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_LABEL

	    DO_ADD_MENU_SEP

            op = EDV_IMBR_OP_DEVICES;
            w_record = &imbr->devices_mi;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_LABEL
        }
        GUIMenuAddToMenuBar(
            menu_bar, menu,
            "Device",
            GUI_MENU_BAR_ALIGN_LEFT
        );


        /* Create settings menu. */
        menu = GUIMenuCreateTearOff();
        if(menu != NULL)
        {
            op = EDV_IMBR_OP_WRITE_PROTECT;
            w_record = &imbr->write_protect_mi;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_LABEL

	    DO_ADD_MENU_SEP

            op = EDV_IMBR_OP_OPTIONS;
            w_record = &imbr->options_mi;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_LABEL

            op = EDV_IMBR_OP_CUSTOMIZE;
            w_record = &imbr->customize_mi;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_LABEL
        }
        GUIMenuAddToMenuBar(
            menu_bar, menu,
            "Settings",
            GUI_MENU_BAR_ALIGN_LEFT
        );


        /* Create windows menu. */
        menu = GUIMenuCreateTearOff();
        if(menu != NULL)
        {
            op = EDV_IMBR_OP_NEW_BROWSER;
            w_record = &imbr->window_new_browser_mi;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_LABEL

            op = EDV_IMBR_OP_NEW_IMBR;
            w_record = &imbr->window_new_imbr_mi;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_LABEL

            op = EDV_IMBR_OP_NEW_ARCHIVER;
            w_record = &imbr->window_new_archiver_mi;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_LABEL

	    DO_ADD_MENU_SEP

            op = EDV_IMBR_OP_RECYCLE_BIN;
            w_record = &imbr->window_recycle_bin_mi;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_LABEL
        }
        GUIMenuAddToMenuBar(
            menu_bar, menu,
            "Windows",
            GUI_MENU_BAR_ALIGN_LEFT
        );





        /* Create help menu. */
        menu = GUIMenuCreateTearOff();
        if(menu != NULL)
        {
            op = EDV_IMBR_OP_HELP_CONTENTS;
            w_record = &imbr->help_contents_mi;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_LABEL

            DO_ADD_MENU_SEP

            op = EDV_IMBR_OP_HELP_KEYS_LIST;
            w_record = &imbr->help_keys_list_mi;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_LABEL

            op = EDV_IMBR_OP_HELP_MIME_TYPES;
            w_record = &imbr->help_mime_types_mi;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_LABEL

            op = EDV_IMBR_OP_HELP_DEVICES;
            w_record = &imbr->help_devices_mi;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_LABEL

            op = EDV_IMBR_OP_HELP_COMMON_OPERATIONS;
            w_record = &imbr->help_common_operations_mi;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_LABEL

            DO_ADD_MENU_SEP

            op = EDV_IMBR_OP_HELP_ABOUT;
            w_record = &imbr->help_about_mi;
            fw_record = NULL;
            DO_ADD_MENU_ITEM_LABEL
        }
        GUIMenuAddToMenuBar(
            menu_bar, menu,
            "Help",
            GUI_MENU_BAR_ALIGN_RIGHT
        );


#undef DO_ADD_MENU_ITEM_LABEL
#undef DO_ADD_MENU_ITEM_SUBMENU
#undef DO_ADD_MENU_ITEM_CHECK
#undef DO_ADD_MENU_SEP

}

/*
 *	Creates a new image browser.
 */
edv_imbr_struct *EDVImbrNew(gpointer core_ptr)
{
	gbool lists_double_buffer = TRUE;
	gint quality = 1;
	gint toplevel_x = 0, toplevel_y = 0;
	gint toplevel_width = 700, toplevel_height = 400;
	gint paned_position = 200;
	gbool	tool_bar_map_state = TRUE,
		location_bar_map_state = TRUE,
		mount_bar_map_state = TRUE,
		find_bar_map_state = TRUE,
		status_bar_map_state = TRUE;
	gbool thumb_list_horizontal = FALSE;
	gint	thumb_width = DEF_IMBR_THUMB_WIDTH,
		thumb_height = DEF_IMBR_THUMB_HEIGHT,
		thumb_border = DEF_IMBR_THUMB_BORDER;
	edv_status_bar_struct *status_bar;
	GdkColormap *colormap = NULL;
	GList *glist;
        GtkWidget *w, *parent, *parent2, *parent3, *parent4;
	gpointer combo_rtn;
	GtkCombo *combo;
        GdkWindow *window;
        GtkAccelGroup *accelgrp;
	tlist_struct *tlist;
	imgview_struct *iv;
        const GtkTargetEntry dnd_tar_types[] = {
{"text/plain",				0,	EDV_DND_TYPE_INFO_TEXT_PLAIN},
{"text/uri-list",			0,	EDV_DND_TYPE_INFO_TEXT_URI_LIST},
{"STRING",				0,	EDV_DND_TYPE_INFO_STRING},
{EDV_DND_TYPE_STRING_RECYCLED_OBJECT,	0,	EDV_DND_TYPE_INFO_RECYCLED_OBJECT},
{EDV_DND_TYPE_STRING_ARCHIVE_OBJECT,	0,	EDV_DND_TYPE_INFO_ARCHIVE_OBJECT}
        };
        const GtkTargetEntry dnd_src_types[] = {
{"text/plain",				0,	EDV_DND_TYPE_INFO_TEXT_PLAIN},
{"text/uri-list",			0,	EDV_DND_TYPE_INFO_TEXT_URI_LIST},
{"STRING",				0,	EDV_DND_TYPE_INFO_STRING}
        };
	edv_imbr_struct *imbr = (edv_imbr_struct *)g_malloc0(
	    sizeof(edv_imbr_struct)
	);
	if(imbr == NULL)
	    return(imbr);


	/* Get values from core structure. */
	if(core_ptr != NULL)
	{
	    edv_core_struct *c_ptr = (edv_core_struct *)core_ptr;

	    switch(EDVCFGItemListGetValueI(
                c_ptr->cfg_list, EDV_CFG_PARM_IMAGE_QUALITY
            ))
	    {
	      case 0:
		quality = 0;
		break;
              case 1:
                quality = 1;
                break;
              case 2:
                quality = 2;
                break;
	    }
	    lists_double_buffer = (gbool)EDVCFGItemListGetValueI(
                c_ptr->cfg_list, EDV_CFG_PARM_LISTS_DOUBLE_BUFFER
            );

	    toplevel_x = EDVCFGItemListGetValueI(
                c_ptr->cfg_list, EDV_CFG_PARM_IMBR_X
            );
            toplevel_y = EDVCFGItemListGetValueI(
                c_ptr->cfg_list, EDV_CFG_PARM_IMBR_Y
            );
	    toplevel_width = EDVCFGItemListGetValueI(
		c_ptr->cfg_list, EDV_CFG_PARM_IMBR_WIDTH
	    );
            toplevel_height = EDVCFGItemListGetValueI(
                c_ptr->cfg_list, EDV_CFG_PARM_IMBR_HEIGHT
            );
            paned_position = EDVCFGItemListGetValueI(
                c_ptr->cfg_list, EDV_CFG_PARM_IMBR_PANED_POSITION
            );

	    tool_bar_map_state = EDVCFGItemListGetValueI(
                c_ptr->cfg_list, EDV_CFG_PARM_IMBR_SHOW_TOOL_BAR
            );
	    location_bar_map_state = EDVCFGItemListGetValueI(
                c_ptr->cfg_list, EDV_CFG_PARM_IMBR_SHOW_LOCATION_BAR
            );
            mount_bar_map_state = EDVCFGItemListGetValueI(
                c_ptr->cfg_list, EDV_CFG_PARM_IMBR_SHOW_MOUNT_BAR
            );
            find_bar_map_state = EDVCFGItemListGetValueI(
                c_ptr->cfg_list, EDV_CFG_PARM_IMBR_SHOW_FIND_BAR
            );
	    status_bar_map_state = EDVCFGItemListGetValueI(
                c_ptr->cfg_list, EDV_CFG_PARM_IMBR_SHOW_STATUS_BAR
            );

	    thumb_list_horizontal = EDVCFGItemListGetValueI(
                c_ptr->cfg_list, EDV_CFG_PARM_IMBR_THUMB_LIST_HORIZONTAL
            );
            thumb_width = EDVCFGItemListGetValueI(
                c_ptr->cfg_list, EDV_CFG_PARM_IMBR_THUMB_WIDTH
            );
            thumb_height = EDVCFGItemListGetValueI(
                c_ptr->cfg_list, EDV_CFG_PARM_IMBR_THUMB_HEIGHT
            );
            thumb_border = EDVCFGItemListGetValueI(
                c_ptr->cfg_list, EDV_CFG_PARM_IMBR_THUMB_BORDER
            );
	}

	/* Begin resetting values. */
	imbr->initialized = TRUE;
	imbr->map_state = FALSE;
        imbr->processing = FALSE;
	imbr->ignore_input = FALSE;
	imbr->busy_count = 0;
        imbr->core_ptr = core_ptr;
	imbr->opid = NULL;
	imbr->total_opids = 0;
	imbr->stop_count = 0;
	imbr->loading_tocb = (guint)-1;
	imbr->tool_bar_map_state = tool_bar_map_state;
	imbr->location_bar_map_state = location_bar_map_state;
        imbr->mount_bar_map_state = mount_bar_map_state;
	imbr->find_bar_map_state = find_bar_map_state;
	imbr->status_bar_map_state = status_bar_map_state;
	imbr->tlist_selected_thumb = -1;
	imbr->selected_dev_num = -1;
	imbr->last_recbin_items = -1;
	imbr->last_write_protect_state = -1;
	imbr->status_bar = NULL;
	imbr->fopdialog = NULL;

	/* Create operation ids. */
	imbr->opid = EDVImbrOPIDListNew(
	    imbr, &imbr->total_opids
	);

        /* Keyboard accelerator group. */
        imbr->accelgrp = accelgrp = gtk_accel_group_new();

	/* Begin creating widgets. */
	/* Toplevel. */
        gtk_widget_push_visual(gdk_rgb_get_visual());
        gtk_widget_push_colormap(gdk_rgb_get_cmap());
	imbr->toplevel = w = gtk_window_new(GTK_WINDOW_TOPLEVEL);
        gtk_widget_pop_visual();
        gtk_widget_pop_colormap();
        gtk_widget_set_uposition(
            w, toplevel_x, toplevel_y
        );
        gtk_widget_set_usize(
            w, toplevel_width, toplevel_height
        );
        gtk_window_set_policy(
            GTK_WINDOW(w),
            TRUE, TRUE, TRUE
        );
	gtk_widget_realize(w);
	window = w->window;
        if(window != NULL)
        {
            GdkGeometry geometry;

            geometry.min_width = 100;
            geometry.min_height = 70;

            geometry.base_width = 0;
            geometry.base_height = 0;

            geometry.width_inc = 1;
            geometry.height_inc = 1;
/*
            geometry.min_aspect = 1.3;
            geometry.max_aspect = 1.3;
 */
            gdk_window_set_geometry_hints(
                window, &geometry,
                GDK_HINT_MIN_SIZE |
                GDK_HINT_BASE_SIZE |
                /* GDK_HINT_ASPECT | */
                GDK_HINT_RESIZE_INC
            );

	    colormap = gdk_window_get_colormap(window);

	    GUISetWMIcon(window, (u_int8_t **)image_browser_48x48_xpm);
        }
	gtk_signal_connect(
	    GTK_OBJECT(w), "delete_event",
	    GTK_SIGNAL_FUNC(EDVImbrDeleteEventCB), imbr
        );
        gtk_signal_connect(
            GTK_OBJECT(w), "destroy",
	    GTK_SIGNAL_FUNC(EDVImbrDestroyCB), imbr
        );
        gtk_accel_group_attach(accelgrp, GTK_OBJECT(w));
        gtk_container_set_border_width(GTK_CONTAINER(w), 0);
        parent = w;


	/* Main vbox. */
	imbr->main_vbox = w = gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(parent), w);
	gtk_widget_show(w);
	parent = w;


	/* Vbox for tool bars. */
	imbr->bars_vbox = w = gtk_vbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
        parent2 = w;


	/* Handle for menu bar. */
        imbr->menu_bar_handle = w = gtk_handle_box_new();
        gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
        gtk_signal_connect_after(
            GTK_OBJECT(w), "child_attached",
            GTK_SIGNAL_FUNC(EDVImbrHandleChildAttachedCB), imbr
        );
        gtk_signal_connect_after(
            GTK_OBJECT(w), "child_detached",
            GTK_SIGNAL_FUNC(EDVImbrHandleChildDetachedCB), imbr
        );
	gtk_widget_show(w);
	/* Create menu bar. */
	EDVImbrCreateMenuBar(imbr, w);


	/* Handle for tool bar. */
	imbr->tool_bar_handle = w = gtk_handle_box_new();
        gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
        gtk_signal_connect_after(
            GTK_OBJECT(w), "child_attached",
            GTK_SIGNAL_FUNC(EDVImbrHandleChildAttachedCB), imbr
        );
        gtk_signal_connect_after(
            GTK_OBJECT(w), "child_detached",
            GTK_SIGNAL_FUNC(EDVImbrHandleChildDetachedCB), imbr
        );
	if(imbr->tool_bar_map_state)
	    gtk_widget_show(w);
	/* Build tool bar later. */
	imbr->toolbar = NULL;


	/* Location bar handle box. */
	imbr->location_bar_handle = w = gtk_handle_box_new();
        gtk_signal_connect_after(
            GTK_OBJECT(w), "child_attached",
            GTK_SIGNAL_FUNC(EDVImbrHandleChildAttachedCB), imbr
        );
        gtk_signal_connect_after(
            GTK_OBJECT(w), "child_detached",
            GTK_SIGNAL_FUNC(EDVImbrHandleChildDetachedCB), imbr
        );
        gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	if(imbr->location_bar_map_state)
	    gtk_widget_show(w);
	parent2 = w;

	/* Location bar hbox for location bar. */
	imbr->location_bar_hbox = w = gtk_hbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_widget_show(w);
	parent2 = w;


	/* Hbox for location icon and combo. */
	w = gtk_hbox_new(FALSE, 0);	/* Don't need spacing between icon
					 * and combo because combo has it
					 * already.
					 */
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent3 = w;

	/* Vbox to orient location icon fixed parent. */
	w = gtk_vbox_new(TRUE, 0);
        gtk_widget_set_usize(w, 20, 20);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent4 = w;
	/* Location icon fixed parent. */
        imbr->location_icon_fixed = w = gtk_fixed_new();
        gtk_box_pack_start(GTK_BOX(parent4), w, TRUE, FALSE, 0);
        gtk_widget_show(w);

	/* Reset location icon pixmap. */
	imbr->location_icon_pm = NULL;

	/* Location combo. */
	glist = NULL;
	w = (GtkWidget *)GUIComboCreate(
#ifdef PROG_LANGUAGE_ENGLISH
	    "Location:",
#endif
#ifdef PROG_LANGUAGE_SPANISH
            "La Ubicacin:",
#endif
#ifdef PROG_LANGUAGE_FRENCH
            "Emplacement:",
#endif
	    "/",
	    glist,
	    20, 	/* Maximum items. */
	    &combo_rtn,
	    imbr,
	    EDVImbrComboActivateCB,
	    NULL
	);
        gtk_box_pack_start(GTK_BOX(parent3), w, TRUE, TRUE, 0);
	gtk_widget_show(w);

	imbr->location_combo = w = (GtkWidget *)combo_rtn;
	combo = GTK_COMBO(w);
        gtk_combo_set_case_sensitive(combo, TRUE);

        EDVEntrySetDND((edv_core_struct *)core_ptr, combo->entry);


        /* Mount bar handle box. */
        imbr->mount_bar_handle = w = gtk_handle_box_new();
        gtk_signal_connect_after(
            GTK_OBJECT(w), "child_attached",
            GTK_SIGNAL_FUNC(EDVImbrHandleChildAttachedCB), imbr
        );
        gtk_signal_connect_after(
            GTK_OBJECT(w), "child_detached",
            GTK_SIGNAL_FUNC(EDVImbrHandleChildDetachedCB), imbr
        );
        gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
        if(imbr->mount_bar_map_state)
            gtk_widget_show(w);
        parent2 = w;

	/* Create mount bar. */
	imbr->mountbar = EDVMountBarNew(
	    core_ptr, parent2,
	    EDVImbrMountBarMountCB,
	    EDVImbrMountBarGoToCB,
	    EDVImbrBarStatusMessageCB,
	    imbr
	);
	EDVMountBarMap(imbr->mountbar);


        /* Find bar handle box. */
        imbr->find_bar_handle = w = gtk_handle_box_new();
        gtk_signal_connect_after(
            GTK_OBJECT(w), "child_attached",
            GTK_SIGNAL_FUNC(EDVImbrHandleChildAttachedCB), imbr
        );
        gtk_signal_connect_after(
            GTK_OBJECT(w), "child_detached",
            GTK_SIGNAL_FUNC(EDVImbrHandleChildDetachedCB), imbr
        );
        gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
        if(imbr->find_bar_map_state)
            gtk_widget_show(w);
        parent2 = w;

        /* Create find bar. */
        imbr->findbar = EDVFindBarNew(
            core_ptr, parent2,
	    EDV_FINDBAR_ROLE_DISK_OBJECT,
            EDVImbrBarStatusMessageCB,
            EDVImbrBarStatusProgressCB,
            EDVImbrFindBarLocationCB,
            EDVImbrFindBarStartCB,
            EDVImbrFindBarEndCB,
            EDVImbrFindBarMatchCB,
            imbr
        );
        EDVFindBarMap(imbr->findbar);



	/* Main paned. */
	imbr->main_paned = w = gtk_hpaned_new();
	gtk_paned_set_position(GTK_PANED(w), paned_position);
	gtk_paned_set_handle_size(
	    GTK_PANED(w), EDV_DEF_PANED_HANDLE_SIZE
	);
	gtk_paned_set_gutter_size(
	    GTK_PANED(w), EDV_DEF_PANED_GUTTER_SIZE
	);
	gtk_box_pack_start(GTK_BOX(parent), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent2 = w;


	/* Create thumbs list. */
	imbr->tlist = tlist = TListNew(
            thumb_list_horizontal,
	    thumb_width, thumb_height, thumb_border,
            imbr,		/* Client data. */
            EDVImbrTListSelectCB, EDVImbrTListUnselectCB
        );
        w = tlist->toplevel;
        gtk_paned_add1(GTK_PANED(parent2), w);
	TListDoubleBuffer(tlist, lists_double_buffer);
	TListSelectionMode(tlist, GTK_SELECTION_EXTENDED);

        w = tlist->list_da;
        gtk_signal_connect(
            GTK_OBJECT(w), "key_press_event",
            GTK_SIGNAL_FUNC(EDVImbrKeyEventCB), imbr
        );
        gtk_signal_connect(
            GTK_OBJECT(w), "key_release_event",
            GTK_SIGNAL_FUNC(EDVImbrKeyEventCB), imbr
        );
        gtk_signal_connect_after(
            GTK_OBJECT(w), "button_press_event",
            GTK_SIGNAL_FUNC(EDVImbrButtonPressEventCB), imbr
        );

        /* Set up DND for the thumbs list. */
        GUIDNDSetSrc(
            w,
            dnd_src_types,
            sizeof(dnd_src_types) / sizeof(GtkTargetEntry),
            GDK_ACTION_COPY | GDK_ACTION_MOVE |
                GDK_ACTION_LINK,                /* Actions. */
            GDK_BUTTON1_MASK | GDK_BUTTON2_MASK,/* Buttons. */
            NULL,
            EDVImbrTListDragDataGetCB,
            EDVImbrTListDragDataDeleteCB,
            NULL,
            imbr
        );
        GUIDNDSetTar(
            w,
            dnd_tar_types,
            sizeof(dnd_tar_types) / sizeof(GtkTargetEntry),
            GDK_ACTION_COPY | GDK_ACTION_MOVE |
                GDK_ACTION_LINK,                /* Actions. */
            GDK_ACTION_MOVE,                    /* Default action if same. */
            GDK_ACTION_MOVE,                    /* Default action. */
            EDVImbrTListDragDataReceivedCB,
            imbr
        );

        TListMap(tlist);


        /* Create image viewer as a child widget. */
	imbr->imgview = iv = ImgViewNew(
            TRUE,			/* Show tool bar. */
            TRUE,			/* Show values. */
            FALSE,			/* Show status bar. */
            FALSE,			/* Show image on wm icon. */
	    quality,			/* Quality (0 to 2). */
	    FALSE,			/* Toplevel to be a GtkWindow? */
            &w				/* Toplevel return. */
        );
        gtk_paned_add2(GTK_PANED(parent2), w);
	ImgViewMap(iv);
        /* Set up view widget. */
        w = (GtkWidget *)ImgViewGetViewWidget(iv);
        if(w != NULL)
        {
            gtk_signal_connect_after(
                GTK_OBJECT(w), "button_press_event",
                GTK_SIGNAL_FUNC(EDVImbrButtonPressEventCB), imbr
            );

            GUIDNDSetTar(
                w,
		dnd_tar_types,
		sizeof(dnd_tar_types) / sizeof(GtkTargetEntry),
		GDK_ACTION_COPY | GDK_ACTION_MOVE |
		    GDK_ACTION_LINK,		/* Actions. */
		GDK_ACTION_MOVE,		/* Default action if same. */
		GDK_ACTION_MOVE,		/* Default action. */
		EDVImbrImgViewDragDataReceivedCB,
		imbr
	    );
        }


	/* Create tool bar. */
        EDVImbrToolbarRegenerate(imbr);

	/* Create tlist right-click menu. */
	EDVImbrTListMenuRegenerate(imbr);


	/* Create status bar. */
	imbr->status_bar = status_bar = EDVStatusBarNew(
	    core_ptr, imbr->main_vbox
	);
	if(status_bar != NULL)
	{
	    if(imbr->status_bar_map_state)
		EDVStatusBarMap(status_bar);
	    else
		EDVStatusBarUnmap(status_bar);
	}

	/* Reset all values. */
	EDVImbrReset(imbr, FALSE);

	/* Update menus. */
	EDVImbrUpdateMenus(imbr);

	return(imbr);
}

/*
 *	Resets all values on the given image browser.
 */
void EDVImbrReset(edv_imbr_struct *imbr, gbool need_unmap)
{
        if(imbr == NULL)
            return;

        if(!imbr->initialized)
            return;


	EDVImbrSetTitle(imbr, NULL);


	if(need_unmap)
	    EDVImbrUnmap(imbr);
}

/*
 *	Syncronizes all data in the image browser's widgets to internal
 *	data on the image browser.
 */
void EDVImbrSyncData(edv_imbr_struct *imbr)
{
        if(imbr == NULL)
            return;

        if(!imbr->initialized)
            return;

/* This function may be removed in the future, since there really is
 * no floating data from widgets to be synced to disk on the image
 * browser.
 */


}


/*
 *	Updates all widgets on the given image browser to reflect the
 *	current data.
 */
void EDVImbrUpdateMenus(edv_imbr_struct *imbr)
{
	gbool write_protect, write_protect_changed = FALSE;
	gbool loading_in_progress;
	gbool sensitive, state;
	gint selected_dev_num;
	edv_device_struct *selected_dev_ptr;
	guint8 **icon_data, **recbin_icon_data;
	GtkWidget *w;
	gint tlist_selected_thumb;
	toolbar_struct *tb;
	edv_core_struct *core_ptr;


	if(imbr == NULL)
	    return;

        if(!imbr->initialized)
            return;

	core_ptr = (edv_core_struct *)imbr->core_ptr;
	if(core_ptr == NULL)
	    return;

        tb = imbr->toolbar;

#define DO_SET_SENSITIVE	\
{ \
 if(w != NULL) \
  gtk_widget_set_sensitive(w, sensitive); \
}
#define DO_SHOW	\
{ \
 if(w != NULL) \
  gtk_widget_show(w); \
}
#define DO_HIDE \
{ \
 if(w != NULL) \
  gtk_widget_hide(w); \
}
#define DO_SET_CHECK_STATE	\
{ \
 if((w != NULL) ? GTK_IS_CHECK_MENU_ITEM(w) : FALSE) \
  GTK_CHECK_MENU_ITEM(w)->active = state; \
}


	/* Get loading in progress state. */
	loading_in_progress = (imbr->loading_tocb != (guint)-1) ? TRUE : FALSE;

	/* Get global write protect state. */
	write_protect = EDVCFGItemListGetValueI(
	    core_ptr->cfg_list, EDV_CFG_PARM_WRITE_PROTECT
	);
	if(imbr->last_write_protect_state < 0)
	{
	    write_protect_changed = TRUE;
	}
	else
	{
	    if((write_protect && !imbr->last_write_protect_state) ||
	       (!write_protect && imbr->last_write_protect_state)
	    )
		write_protect_changed = TRUE;
	}
	imbr->last_write_protect_state = write_protect ? 1 : 0;

	/* Get last selected thumb. */
	tlist_selected_thumb = imbr->tlist_selected_thumb;

	/* Get last selected device (if any). */
	selected_dev_num = imbr->selected_dev_num;
	if((selected_dev_num >= 0) && (selected_dev_num < core_ptr->total_devices))
	    selected_dev_ptr = core_ptr->device[selected_dev_num];
	else
	    selected_dev_ptr = NULL;

	/* Update recycle bin icon data, setting it to not NULL if it
	 * needs to be changed and update the image browser's
	 * last_recbin_items to match the value on the core structure.
	 */
        icon_data = NULL;
        if(imbr->last_recbin_items < 0)
        {
            if(core_ptr->last_recbin_items > 0)
                icon_data = (guint8 **)icon_trash_20x20_xpm;
            else
                icon_data = (guint8 **)icon_trash_empty_20x20_xpm;
        }
        else
        {
            if(!imbr->last_recbin_items &&
               core_ptr->last_recbin_items
            )
                icon_data = (guint8 **)icon_trash_20x20_xpm;
            else if(imbr->last_recbin_items &&
                    !core_ptr->last_recbin_items
            )
                icon_data = (guint8 **)icon_trash_empty_20x20_xpm;
        }
        imbr->last_recbin_items = core_ptr->last_recbin_items;
	recbin_icon_data = icon_data;



	/* Update mount bar. */
	EDVMountBarUpdateMenus(imbr->mountbar);

        /* Update find bar. */
        EDVFindBarUpdateMenus(imbr->findbar);



        /* Begin updating menu items and tool bar items. */

        /* File. */
        /* Open and open with. */
        sensitive = (tlist_selected_thumb > -1) ? TRUE : FALSE;
        /* Open. */
	w = imbr->open_mi;
	DO_SET_SENSITIVE
        ToolBarItemSetSensitiveID(
            tb, EDV_IMBR_OP_OPEN, sensitive
        );
        /* Open with. */
        w = imbr->openwith_mi;
        DO_SET_SENSITIVE
        ToolBarItemSetSensitiveID(
            tb, EDV_IMBR_OP_OPEN_WITH, sensitive
        );

        /* New object sub menu item. */
	sensitive = TRUE;
        w = imbr->new_object_submi;
	DO_SET_SENSITIVE
        /* New file. */
        w = imbr->new_file_mi;
        DO_SET_SENSITIVE
        ToolBarItemSetSensitiveID(
            tb, EDV_IMBR_OP_NEW_FILE, sensitive
        );
	/* New directory. */
        w = imbr->new_directory_mi;
        DO_SET_SENSITIVE
        ToolBarItemSetSensitiveID(
            tb, EDV_IMBR_OP_NEW_DIRECTORY, sensitive
        );
	/* New link. */
        w = imbr->new_link_mi;
        DO_SET_SENSITIVE
        ToolBarItemSetSensitiveID(
            tb, EDV_IMBR_OP_NEW_LINK, sensitive
        );
	/* New fifo. */
        w = imbr->new_fifo_mi;
        DO_SET_SENSITIVE
        ToolBarItemSetSensitiveID(
            tb, EDV_IMBR_OP_NEW_FIFO, sensitive
        );
	/* New block device. */
        w = imbr->new_device_block_mi;
        DO_SET_SENSITIVE
        ToolBarItemSetSensitiveID(
            tb, EDV_IMBR_OP_NEW_DEVICE_BLOCK, sensitive
        );
	/* New character device. */
        w = imbr->new_device_character_mi;
        DO_SET_SENSITIVE
        ToolBarItemSetSensitiveID(
            tb, EDV_IMBR_OP_NEW_DEVICE_CHARACTER, sensitive
        );
	/* New socket. */
        w = imbr->new_socket_mi;
        DO_SET_SENSITIVE
        ToolBarItemSetSensitiveID(
            tb, EDV_IMBR_OP_NEW_SOCKET, sensitive
        );


	/* Edit. */
        sensitive = (tlist_selected_thumb > -1) ? TRUE : FALSE;
        /* Move. */
        w = imbr->move_mi;
        DO_SET_SENSITIVE
        ToolBarItemSetSensitiveID(
            tb, EDV_IMBR_OP_MOVE, sensitive
        );
	/* Copy. */
        w = imbr->copy_mi;
        DO_SET_SENSITIVE
        ToolBarItemSetSensitiveID(
            tb, EDV_IMBR_OP_COPY, sensitive
        );
	/* Link. */
        w = imbr->link_mi;
        DO_SET_SENSITIVE
        ToolBarItemSetSensitiveID(
            tb, EDV_IMBR_OP_LINK, sensitive
        );
	/* Rename. */
        w = imbr->rename_mi;
        DO_SET_SENSITIVE
        ToolBarItemSetSensitiveID(
            tb, EDV_IMBR_OP_RENAME, sensitive
        );
	/* Change mode. */
        w = imbr->chmod_mi;
        DO_SET_SENSITIVE
        ToolBarItemSetSensitiveID(
            tb, EDV_IMBR_OP_CHMOD, sensitive
        );
	/* Chown. */
        w = imbr->chown_mi;
        DO_SET_SENSITIVE
        ToolBarItemSetSensitiveID(
            tb, EDV_IMBR_OP_CHOWN, sensitive
        );
	/* Delete. */
        w = imbr->delete_mi;
        DO_SET_SENSITIVE
        ToolBarItemSetSensitiveID(
            tb, EDV_IMBR_OP_DELETE, sensitive
        );
	/* Properties. */
        w = imbr->properties_mi;
        DO_SET_SENSITIVE
        ToolBarItemSetSensitiveID(
            tb, EDV_IMBR_OP_PROPERTIES, sensitive
        );


	/* View. */
	/* Stop loading. */
	sensitive = loading_in_progress;
	w = imbr->stop_mi;
	DO_SET_SENSITIVE
        ToolBarItemSetSensitiveID(
            tb, EDV_IMBR_OP_STOP, sensitive
	);
	/* Continue loading. */
	sensitive = !loading_in_progress;
        w = imbr->continue_mi;
        DO_SET_SENSITIVE
        ToolBarItemSetSensitiveID(
            tb, EDV_IMBR_OP_CONTINUE, sensitive
	);

        /* Show tool bar. */
	w = imbr->show_tool_bar_micheck;
	state = EDVCFGItemListGetValueI(
            core_ptr->cfg_list, EDV_CFG_PARM_IMBR_SHOW_TOOL_BAR
	);
	DO_SET_CHECK_STATE
        ToolBarItemSetToggleID(
            tb, EDV_IMBR_OP_SHOW_TOOL_BAR, state
        );
        /* Show location bar. */
        w = imbr->show_location_bar_micheck;
        state = EDVCFGItemListGetValueI(
            core_ptr->cfg_list, EDV_CFG_PARM_IMBR_SHOW_LOCATION_BAR
        );
        DO_SET_CHECK_STATE
        ToolBarItemSetToggleID(
            tb, EDV_IMBR_OP_SHOW_LOCATION_BAR, state
        );
        /* Show mount bar. */
        w = imbr->show_mount_bar_micheck;
        state = EDVCFGItemListGetValueI(
            core_ptr->cfg_list, EDV_CFG_PARM_IMBR_SHOW_MOUNT_BAR
        );
        DO_SET_CHECK_STATE
        ToolBarItemSetToggleID(
            tb, EDV_IMBR_OP_SHOW_MOUNT_BAR, state
        );
        /* Show find bar. */
        w = imbr->show_find_bar_micheck;
        state = EDVCFGItemListGetValueI(
            core_ptr->cfg_list, EDV_CFG_PARM_IMBR_SHOW_FIND_BAR
        );
        DO_SET_CHECK_STATE
        ToolBarItemSetToggleID(
            tb, EDV_IMBR_OP_SHOW_FIND_BAR, state
        );
        /* Show status bar. */
        w = imbr->show_status_bar_micheck;
        state = EDVCFGItemListGetValueI(
            core_ptr->cfg_list, EDV_CFG_PARM_IMBR_SHOW_STATUS_BAR
        );
        DO_SET_CHECK_STATE
        ToolBarItemSetToggleID(
            tb, EDV_IMBR_OP_SHOW_STATUS_BAR, state
        );


	/* Device. */
	w = imbr->mount_mi;
	sensitive = (selected_dev_ptr != NULL) ?
	    !selected_dev_ptr->no_unmount : FALSE;
	DO_SET_SENSITIVE
        if(1)
        {
            const gchar *text;

            if(selected_dev_ptr != NULL)
                text =  selected_dev_ptr->is_mounted ?
                    "Unmount" : "Mount";
            else
                text = "Mount/Unmount";
            GUIMenuItemSetLabel(w, text);
            ToolBarItemUpdateByID(
                tb, EDV_IMBR_OP_MOUNT, text, NULL, NULL
            );
        }
        ToolBarItemSetSensitiveID(
            tb, EDV_IMBR_OP_MOUNT, sensitive
        );


	/* Settings. */
        if(write_protect_changed)
	{
	    w = imbr->write_protect_mi;
	    if(w != NULL)
		GUIMenuItemSetPixmap(
		    w,
		    (u_int8_t **)(write_protect ?
			icon_secure_20x20_xpm : icon_insecure_20x20_xpm)
		);
            ToolBarItemUpdateByID(
                tb, EDV_IMBR_OP_WRITE_PROTECT,
                NULL,
                (guint8 **)(write_protect ?
                    icon_secure_20x20_xpm : icon_insecure_20x20_xpm),
                NULL
            );
	}


	/* Windows. */
	if(recbin_icon_data != NULL)
	{
	    GUIMenuItemSetPixmap(
		imbr->window_recycle_bin_mi, recbin_icon_data
	    );
            ToolBarItemUpdateByID(
                tb, EDV_IMBR_OP_RECYCLE_BIN,
                NULL, recbin_icon_data, NULL
            );
	}


	/* Update right-click menu. */
        sensitive = (tlist_selected_thumb > -1) ? TRUE : FALSE;
	w = imbr->tlist_open_mi;
        DO_SET_SENSITIVE
        w = imbr->tlist_openwith_mi;
        DO_SET_SENSITIVE
        w = imbr->tlist_move_mi;
        DO_SET_SENSITIVE
        w = imbr->tlist_copy_mi;
        DO_SET_SENSITIVE
        w = imbr->tlist_link_mi;
        DO_SET_SENSITIVE
        w = imbr->tlist_rename_mi;
        DO_SET_SENSITIVE
        w = imbr->tlist_chmod_mi;
        DO_SET_SENSITIVE
        w = imbr->tlist_chown_mi;
        DO_SET_SENSITIVE
        w = imbr->tlist_delete_mi;
        DO_SET_SENSITIVE
        w = imbr->tlist_properties_mi;
        DO_SET_SENSITIVE

	w = imbr->tlist_mount_mi;
	if(selected_dev_ptr == NULL)
	    DO_HIDE
	else
	    DO_SHOW
        sensitive = (selected_dev_ptr != NULL) ?
            !selected_dev_ptr->no_unmount : FALSE;
        DO_SET_SENSITIVE
        if(selected_dev_ptr != NULL)
        {
            const gchar *text = selected_dev_ptr->is_mounted ?
		"Unmount" : "Mount";
            GUIMenuItemSetLabel(w, text);
        }

        w = imbr->tlist_mount_misep;
        if(selected_dev_ptr == NULL)
            DO_HIDE
        else
            DO_SHOW







	/* Update status bar. */
        EDVStatusBarUpdateMenus(imbr->status_bar);

#undef DO_SHOW
#undef DO_HIDE
#undef DO_SET_SENSITIVE
#undef DO_SET_CHECK_STATE
}

/*
 *	Marks the image browser as busy or ready.
 */
void EDVImbrSetBusy(edv_imbr_struct *imbr, gbool is_busy)
{
	GdkCursor *cursor;
        GtkWidget *w;
	edv_core_struct *core_ptr;


        if(imbr == NULL)
            return;

        if(!imbr->initialized)
            return;

	core_ptr = (edv_core_struct *)imbr->core_ptr;
	if(core_ptr == NULL)
	    return;

	w = imbr->toplevel;
	if(w != NULL)
	{
	    if(is_busy)
	    {
		/* Increase busy count. */
		imbr->busy_count++;

		/* If already busy then don't change anything. */
		if(imbr->busy_count > 1)
		    return;

		cursor = EDVGetCursor(core_ptr, EDV_CURSOR_CODE_BUSY);
	    }
	    else
	    {
		/* Reduce busy count. */
		imbr->busy_count--;
		if(imbr->busy_count < 0)
		    imbr->busy_count = 0;

		/* If still busy do not change anything. */
		if(imbr->busy_count > 0)
                    return;

		cursor = NULL;	/* Use default cursor. */
	    }

	    /* Update toplevel window's cursor. */
	    if(w->window != NULL)
	    {
		gdk_window_set_cursor(w->window, cursor);
		gdk_flush();
	    }
	}
}


/*
 *	Maps the given image browser.
 */
void EDVImbrMap(edv_imbr_struct *imbr)
{
        GtkWidget *w;


        if(imbr == NULL)
            return;

        if(!imbr->initialized)
            return;

        if(!imbr->map_state)
        {
            w = imbr->toplevel;
            if(w != NULL)
                gtk_widget_show(w);

            imbr->map_state = TRUE;
        }
}

/*
 *	Unmaps the given image browser.
 */
void EDVImbrUnmap(edv_imbr_struct *imbr)
{
        GtkWidget *w;


        if(imbr == NULL)
            return;

        if(!imbr->initialized)
            return;

        if(imbr->map_state)
        {
            w = imbr->toplevel;
            if(w != NULL)
                gtk_widget_hide(w);

            imbr->map_state = FALSE;
        }
}

/*
 *	Destroys all resources on the given image browser and deallocates
 *	the structure itself.
 */
void EDVImbrDelete(edv_imbr_struct *imbr)
{
        GtkWidget **w;


	if(imbr == NULL)
	    return;

	if(imbr->initialized)
	{
#define DO_UNREF_STYLE		\
{ \
 if((*style) != NULL) \
 { \
  GtkStyle *ts = *style; \
  (*style) = NULL; \
  gtk_style_unref(ts); \
 } \
}

#define DO_DESTROY_WIDGET	\
{ \
 if((*w) != NULL) \
 { \
  GtkWidget *tmp_w = *w; \
  (*w) = NULL; \
  gtk_widget_destroy(tmp_w); \
 } \
}

	    /* Stop timeout. */
	    if(imbr->loading_tocb != (guint)-1)
	    {
		gtk_timeout_remove(imbr->loading_tocb);
		imbr->loading_tocb = (guint)-1;
	    }


	    /* Reset last selected item references. */
	    imbr->tlist_selected_thumb = -1;

	    /* Delete file operation dialog. */
	    EDVFOPDialogDelete(imbr->fopdialog);
	    imbr->fopdialog = NULL;


	    /* Right click menu. */
	    w = &imbr->tlist_menu;
	    imbr->tlist_open_mi = NULL;
            imbr->tlist_openwith_mi = NULL;
            imbr->tlist_mount_mi = NULL;
            imbr->tlist_mount_misep = NULL;
            imbr->tlist_move_mi = NULL;
            imbr->tlist_copy_mi = NULL;
            imbr->tlist_link_mi = NULL;
            imbr->tlist_rename_mi = NULL;
            imbr->tlist_chmod_mi = NULL;
            imbr->tlist_chown_mi = NULL;
            imbr->tlist_delete_mi = NULL;
            imbr->tlist_properties_mi = NULL;
	    DO_DESTROY_WIDGET


	    /* Delete status bar. */
	    EDVStatusBarDelete(imbr->status_bar);
	    imbr->status_bar = NULL;

	    /* Delete tool bar. */
	    ToolBarDelete(imbr->toolbar);
	    imbr->toolbar = NULL;

	    /* Delete mount bar. */
	    EDVMountBarDelete(imbr->mountbar);
	    imbr->mountbar = NULL;

            /* Delete find bar. */
            EDVFindBarDelete(imbr->findbar);
            imbr->findbar = NULL;

	    /* Destroy location bar widgets. */
	    w = &imbr->location_icon_pm;
	    DO_DESTROY_WIDGET
	    w = &imbr->location_icon_fixed;
	    DO_DESTROY_WIDGET
	    w = &imbr->location_combo;
	    DO_DESTROY_WIDGET
	    w = &imbr->location_bar_hbox;
	    DO_DESTROY_WIDGET


	    /* Destroy menu bar and submenus. */
	    w = &imbr->new_object_submenu;
            imbr->new_file_mi = NULL;
            imbr->new_directory_mi = NULL;
            imbr->new_link_mi = NULL;
            imbr->new_fifo_mi = NULL;
            imbr->new_device_block_mi = NULL;
            imbr->new_device_character_mi = NULL;
            imbr->new_socket_mi = NULL;
	    DO_DESTROY_WIDGET

	    w = &imbr->menu_bar;
            imbr->open_mi = NULL;
            imbr->openwith_mi = NULL;
	    imbr->new_object_submi = NULL;
            imbr->sync_disks_mi = NULL;
            imbr->close_mi = NULL;

	    imbr->move_mi = NULL;
	    imbr->copy_mi = NULL;
	    imbr->link_mi = NULL;
            imbr->rename_mi = NULL;
            imbr->chmod_mi = NULL;
            imbr->chown_mi = NULL;
            imbr->delete_mi = NULL;
            imbr->select_all_mi = NULL;
            imbr->unselect_all_mi = NULL;
            imbr->find_mi = NULL;
	    imbr->properties_mi = NULL;

	    imbr->mount_mi = NULL;
	    imbr->devices_mi = NULL;

	    imbr->stop_mi = NULL;
	    imbr->continue_mi = NULL;
	    imbr->refresh_mi = NULL;
	    imbr->refresh_all_mi = NULL;
            imbr->show_tool_bar_micheck = NULL;
            imbr->show_location_bar_micheck = NULL;
            imbr->show_mount_bar_micheck = NULL;
            imbr->show_find_bar_micheck = NULL;
            imbr->show_status_bar_micheck = NULL;
            imbr->mime_types_mi = NULL;

            imbr->write_protect_mi = NULL;
            imbr->options_mi = NULL;
            imbr->customize_mi = NULL;

            imbr->window_new_browser_mi = NULL;
	    imbr->window_new_imbr_mi = NULL;
            imbr->window_new_archiver_mi = NULL;
	    imbr->window_recycle_bin_mi = NULL;

            imbr->help_contents_mi = NULL;
	    imbr->help_keys_list_mi = NULL;
            imbr->help_mime_types_mi = NULL;
            imbr->help_devices_mi = NULL;
            imbr->help_common_operations_mi = NULL;
            imbr->help_about_mi = NULL;

	    DO_DESTROY_WIDGET

	    /* Destroy thumbs list and image viewer. */
	    TListDelete(imbr->tlist);
	    imbr->tlist = NULL;
	    ImgViewDelete(imbr->imgview);
	    imbr->imgview = NULL;


	    /* Destroy toplevel and the rest of the widgets. */
            w = &imbr->toplevel;
            imbr->main_vbox = NULL;
	    imbr->menu_bar = NULL;
	    imbr->bars_vbox = NULL;
            imbr->location_bar_handle = NULL;
            imbr->tool_bar_handle = NULL;
            imbr->menu_bar_handle = NULL;
	    imbr->main_paned = NULL;
            DO_DESTROY_WIDGET

            if(imbr->accelgrp != NULL)
            {
                gtk_accel_group_unref(imbr->accelgrp);
                imbr->accelgrp = NULL;
            }

#undef DO_UNREF_STYLE
#undef DO_DESTROY_WIDGET
	}


	/* Deallocate other memory on image browser structure. */

	/* Operation id structures. */
	EDVImbrOPIDListDelete(
	    imbr->opid, imbr->total_opids
	);



	/* Deallocate structure itself. */
	g_free(imbr);
}
