/*
                      Endeavour Image Browser
 */

#ifndef IMBR_H
#define IMBR_H


#include <gtk/gtk.h>

#include "edvtypes.h"
#include "edvstatusbar.h"
#include "toolbar.h"
#include "mountbar.h"
#include "findbar.h"
#include "fopdialog.h"
#include "tlist.h"
#include "imgview.h"


/*
 *	Image browser operation codes:
 */
#define EDV_IMBR_OP_NONE		0
#define EDV_IMBR_OP_SEPARATOR		10

#define EDV_IMBR_OP_CLOSE		20
#define EDV_IMBR_OP_EXIT		21

#define EDV_IMBR_OP_SYNC_DISKS		30
#define EDV_IMBR_OP_WRITE_PROTECT	35

#define EDV_IMBR_OP_NEW			40
#define EDV_IMBR_OP_NEW_FILE		41
#define EDV_IMBR_OP_NEW_DIRECTORY	42
#define EDV_IMBR_OP_NEW_LINK		43
#define EDV_IMBR_OP_NEW_FIFO		44
#define EDV_IMBR_OP_NEW_DEVICE_BLOCK	45
#define EDV_IMBR_OP_NEW_DEVICE_CHARACTER	46
#define EDV_IMBR_OP_NEW_SOCKET		47

#define EDV_IMBR_OP_OPEN		50
#define EDV_IMBR_OP_OPEN_WITH		51

#define EDV_IMBR_OP_GOTO_PARENT		60
#define EDV_IMBR_OP_MOUNT		61

#define EDV_IMBR_OP_MOVE		65
#define EDV_IMBR_OP_COPY		66
#define EDV_IMBR_OP_LINK		67
#define EDV_IMBR_OP_RENAME		68
#define EDV_IMBR_OP_CHMOD		69
#define EDV_IMBR_OP_CHOWN		70
#define EDV_IMBR_OP_DELETE		71
#define EDV_IMBR_OP_SELECT_ALL		72
#define EDV_IMBR_OP_UNSELECT_ALL	73
#define EDV_IMBR_OP_PROPERTIES		74

#define EDV_IMBR_OP_FIND		75

#define EDV_IMBR_OP_REFRESH		80
#define EDV_IMBR_OP_REFRESH_ALL		81
#define EDV_IMBR_OP_STOP		82
#define EDV_IMBR_OP_CONTINUE		83
#define EDV_IMBR_OP_SHOW_TOOL_BAR	90
#define EDV_IMBR_OP_SHOW_LOCATION_BAR	91
#define EDV_IMBR_OP_SHOW_MOUNT_BAR	92
#define EDV_IMBR_OP_SHOW_FIND_BAR	93
#define EDV_IMBR_OP_SHOW_STATUS_BAR	94
#define EDV_IMBR_OP_MIME_TYPES		100
#define EDV_IMBR_OP_DEVICES		101

#define EDV_IMBR_OP_NEW_BROWSER		110
#define EDV_IMBR_OP_NEW_IMBR		111
#define EDV_IMBR_OP_NEW_ARCHIVER	112
#define EDV_IMBR_OP_RECYCLE_BIN		113

#define EDV_IMBR_OP_OPTIONS		150
#define EDV_IMBR_OP_CUSTOMIZE		151


#define EDV_IMBR_OP_HELP_ABOUT		1000
#define EDV_IMBR_OP_HELP_CONTENTS	1001
#define EDV_IMBR_OP_HELP_KEYS_LIST	1002
#define EDV_IMBR_OP_HELP_MIME_TYPES	1003
#define EDV_IMBR_OP_HELP_DEVICES	1004
#define EDV_IMBR_OP_HELP_COMMON_OPERATIONS	1005


/*
 *	Image browser operation id structure:
 */
typedef struct {

	gpointer imbr;		/* Pointer back to image browser. */

	gint op;		/* One of EDV_IMBR_OP_*. */

	gbool allow_multiple;	/* Allow multiple instances of this
				 * operation in a list if TRUE.
				 */

        gint toolbar_item_type; /* Type of item when placed on the
                                 * tool bar, one of TOOLBAR_ITEM_*.
                                 */

	/* Operation function callback, sole argument is the pointer
	 * to this structure.
	 */
	void (*func_cb)(gpointer);

	/* Enter and leave notify callback, sole argument is the pointer
	 * to this structure.
	 */
	void (*enter_func_cb)(gpointer);
	void (*leave_func_cb)(gpointer);

	gchar *button_name;	/* Name as it would appear on a button. */
	gchar *menu_name;	/* Name as it would appear on a menu item. */
	gchar *tooltip;		/* Tool tip or description (one line). */

	gint accel_key;		/* GTK accelerator key. */
	guint accel_mods;	/* GTK key modifiers for accelerator key. */

	guint8	**button_icon_data,	/* Shared. */
		**menu_icon_data;	/* Shared. */

} edv_imbr_opid_struct;


/*
 *	Image browser structure:
 */
typedef struct {

	gbool initialized;
	gbool map_state;
	gbool processing;
	gbool ignore_input;
	gint busy_count;
	gpointer core_ptr;

	edv_imbr_opid_struct **opid;
	gint total_opids;

	gint stop_count;
	guint loading_tocb;

	GtkAccelGroup *accelgrp;

	gbool	tool_bar_map_state,
		location_bar_map_state,
                mount_bar_map_state,
		find_bar_map_state,
		status_bar_map_state;

	GtkWidget	*toplevel,
			*main_vbox,
			*menu_bar,
			*bars_vbox,		/* Toolbars GtkVBox. */

			*menu_bar_handle,
			*tool_bar_handle,
			*location_bar_handle,
                        *mount_bar_handle,
			*find_bar_handle,

			*location_bar_hbox,	/* GtkHBox parent for location bar. */
			*location_icon_fixed,
			*location_icon_pm,
			*location_combo,

			*main_paned;		/* GtkHPaned */

	/* General tool bar. */
	toolbar_struct *toolbar;

	/* Mount bar. */
	edv_mountbar_struct *mountbar;

        /* Find bar. */
        edv_findbar_struct *findbar;

	/* Important menu items. */
	GtkWidget	*open_mi,
			*openwith_mi,
			*new_object_submi,
			*new_object_submenu,
			*new_file_mi,
			*new_directory_mi,
                        *new_link_mi,
			*new_fifo_mi,
			*new_device_block_mi,
			*new_device_character_mi,
			*new_socket_mi,
			*sync_disks_mi,
			*close_mi,

			*move_mi,
			*copy_mi,
			*link_mi,
                        *rename_mi,
                        *chmod_mi,
                        *chown_mi,
                        *delete_mi,
			*select_all_mi,
			*unselect_all_mi,
			*find_mi,
			*properties_mi,

			*mount_mi,
			*devices_mi,

			*stop_mi,
			*continue_mi,
			*refresh_mi,
			*refresh_all_mi,
                        *show_tool_bar_micheck,
                        *show_location_bar_micheck,
			*show_mount_bar_micheck,
			*show_find_bar_micheck,
                        *show_status_bar_micheck,
			*mime_types_mi,

                        *write_protect_mi,
			*options_mi,
			*customize_mi,

			*window_new_browser_mi,
                        *window_new_imbr_mi,
                        *window_new_archiver_mi,
			*window_recycle_bin_mi,

			*help_contents_mi,
			*help_keys_list_mi,
			*help_mime_types_mi,
			*help_devices_mi,
			*help_common_operations_mi,
			*help_about_mi;

	/* Thumbs list. */
	tlist_struct *tlist;

	/* Image viewer. */
	imgview_struct *imgview;


	/* Right-click menu. */
	GtkWidget	*tlist_menu,
			*tlist_open_mi,
                        *tlist_openwith_mi,
			*tlist_mount_mi,
			*tlist_mount_misep,
			*tlist_move_mi,
                        *tlist_copy_mi,
                        *tlist_link_mi,
                        *tlist_rename_mi,
                        *tlist_chmod_mi,
                        *tlist_chown_mi,
                        *tlist_delete_mi,
                        *tlist_properties_mi;

	/* Last selected thumb, can be -1. Note that this does not reflect
	 * other selected rows since multiple selections are allowed.
	 */
	gint tlist_selected_thumb;

	/* Selected device, reflects the device of the disk object of
	 * the selected thumb. The device number corresponds with a device
	 * on the core's list of devices.
	 */
	gint selected_dev_num;


        /* Last known items in recycle bin, this is used as a quick access
         * value to check if the recycle bin has contents or not. Can be
	 * -1 to indicate undefined.
         */
        gint last_recbin_items;

	/* Last known write protect state, can be -1 to indicate
	 * undefined.
	 */
	gint last_write_protect_state;



	/* Status bar. */
	edv_status_bar_struct *status_bar;


	/* File operation dialog. */
	edv_fopdialog_struct *fopdialog;

} edv_imbr_struct;



extern void EDVImbrSetTitle(edv_imbr_struct *imbr, const gchar *path);

extern gchar *EDVImbrCurrentLocation(edv_imbr_struct *imbr);
extern void EDVImbrSetLocation(
        edv_imbr_struct *imbr,
        const gchar *path, gbool record_history
);
extern void EDVImbrUpdateLocationIcon(
        edv_imbr_struct *imbr, const gchar *path
);

extern void EDVImbrToolbarRegenerate(edv_imbr_struct *imbr);
extern void EDVImbrTListMenuRegenerate(edv_imbr_struct *imbr);

extern void EDVImbrSyncConfiguration(edv_imbr_struct *imbr);

extern void EDVImbrQueueLoadingProcess(edv_imbr_struct *imbr);

extern void EDVImbrSelectPath(
	edv_imbr_struct *imbr, const gchar *path
);

extern edv_imbr_opid_struct **EDVImbrOPIDListNew(
        edv_imbr_struct *imbr,		/* Can be NULL. */
        gint *total
);
extern void EDVImbrOPIDListDelete(
        edv_imbr_opid_struct **list, gint total
);

extern edv_imbr_struct *EDVImbrNew(gpointer core_ptr);
extern void EDVImbrReset(edv_imbr_struct *imbr, gbool need_unmap);
extern void EDVImbrSyncData(edv_imbr_struct *imbr);
extern void EDVImbrUpdateMenus(edv_imbr_struct *imbr);
extern void EDVImbrSetBusy(edv_imbr_struct *imbr, gbool is_busy);
extern void EDVImbrMap(edv_imbr_struct *imbr);
extern void EDVImbrUnmap(edv_imbr_struct *imbr);
extern void EDVImbrDelete(edv_imbr_struct *imbr);


#endif	/* IMBR_H */
